from __future__ import annotations

import os
import random
import tempfile
from dataclasses import dataclass
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest
from typing_extensions import TypeVar

import resonate
from resonate.commands import Command
from resonate.contants import ENV_VARIABLE_PIN_SEED
from resonate.dst.scheduler import DSTScheduler
from resonate.events import (
    ExecutionAwaited,
    ExecutionInvoked,
    ExecutionResumed,
    ExecutionTerminated,
    PromiseCompleted,
    PromiseCreated,
)
from resonate.options import LOptions
from resonate.result import Ok
from resonate.retry_policy import never
from resonate.storage.local_store import (
    LocalStore,
    MemoryStorage,
)
from resonate.storage.resonate_server import RemoteServer
from resonate.testing import dst
from resonate.time import now

if TYPE_CHECKING:
    from collections.abc import Generator

    from resonate.context import Context
    from resonate.dependency_injection import Dependencies
    from resonate.dst.scheduler import DSTScheduler
    from resonate.promise import Promise
    from resonate.storage.traits import IPromiseStore
    from resonate.typing import Yieldable

T = TypeVar("T")


def number(ctx: Context, n: int) -> int:  # noqa: ARG001
    return n


def failing_function(ctx: Context) -> None:  # noqa: ARG001
    msg = "An error happenned"
    raise RuntimeError(msg)


def coro_that_fails_call(ctx: Context) -> Generator[Yieldable, Any, None]:
    return (yield ctx.lfc(failing_function).with_options(retry_policy=never()))


def coro_that_fails_invoke(ctx: Context) -> Generator[Yieldable, Any, None]:
    return (yield (yield ctx.lfi(failing_function).with_options(retry_policy=never())))


def raise_inmediately(ctx: Context) -> Generator[Yieldable, Any, int]:  # noqa: ARG001
    msg = "First thing we do is fail."
    raise RuntimeError(msg)


def only_call(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
    x: int = yield ctx.lfc(number, n=n)
    return x


def only_invocation(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
    xp: Promise[int] = yield ctx.lfi(number, n=n)
    x: int = yield xp
    return x


def failing_asserting(ctx: Context) -> Generator[Yieldable, Any, int]:
    x: int = yield ctx.lfc(only_invocation, n=3)
    ctx.assert_statement(x < 0, f"{x} should be negative")
    return x


def _promise_result(promises: list[Promise[T]]) -> list[T]:
    return [x.result() for x in promises]


def mocked_number() -> int:
    return 23


@dataclass(frozen=True)
class GreetCommand(Command):
    name: str


def batch_greeting(cmds: list[GreetCommand]) -> list[str]:
    return [f"Hello {cmd.name}" for cmd in cmds]


def greet_with_batching(ctx: Context, name: str) -> Generator[Yieldable, Any, str]:
    p: Promise[str] = yield ctx.lfi(GreetCommand(name=name))
    g: str = yield p
    return g


def greet_with_batching_but_with_call(
    ctx: Context, name: str
) -> Generator[Yieldable, Any, str]:
    g: str = yield ctx.lfc(GreetCommand(name=name))
    return g


@cache
def _promise_storages() -> list[IPromiseStore]:
    stores: list[IPromiseStore] = [LocalStore(MemoryStorage())]
    if os.getenv("RESONATE_STORE_URL") is not None:
        stores.append(RemoteServer(url=os.environ["RESONATE_STORE_URL"]))
    return stores


@pytest.mark.parametrize("store", _promise_storages())
def test_raise_inmediately(store: IPromiseStore) -> None:
    s = dst(seeds=[1], durable_promise_storage=store)[0]
    s.add(
        "raise-inmediately-1",
        LOptions(durable=True, retry_policy=never()),
        raise_inmediately,
    )
    p = s.run()[0]
    assert p.failure()
    with pytest.raises(RuntimeError):
        p.result()


@pytest.mark.parametrize("store", _promise_storages())
def test_failing_call(store: IPromiseStore) -> None:
    s = dst(seeds=[1], durable_promise_storage=store)[0]
    s.add(
        "failing-call-1",
        LOptions(durable=True, retry_policy=never()),
        coro_that_fails_call,
    )
    promises = s.run()
    assert (p.failure() for p in promises)
    assert (not p.success() for p in promises)

    s = dst(seeds=[1])[0]
    s.add(
        "failing-invoke-1",
        LOptions(durable=True, retry_policy=never()),
        coro_that_fails_invoke,
    )
    promises = s.run()
    assert (p.failure() for p in promises)
    assert (not p.success() for p in promises)


@pytest.mark.skip
@pytest.mark.parametrize("store", _promise_storages())
def test_batching_using_call(store: IPromiseStore) -> None:
    s = dst(seeds=[1], durable_promise_storage=store)[0]
    s.register_command(cmd=GreetCommand, handler=batch_greeting, max_batch=2)

    s.add(
        "batching-using-call-1",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Ging",
    )
    s.add(
        "batching-using-call-2",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Razor",
    )
    s.add(
        "batching-using-call-3",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching_but_with_call,
        name="Eta",
    )
    s.add(
        "batching-using-call-4",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Elena",
    )
    s.add(
        "batching-using-call-5",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Dwun",
    )
    greetings_promises = s.run()
    assert all(p.success() for p in greetings_promises)
    assert [p.result() for p in greetings_promises] == [
        "Hello Ging",
        "Hello Razor",
        "Hello Eta",
        "Hello Elena",
        "Hello Dwun",
    ]


@pytest.mark.skip
@pytest.mark.parametrize("store", _promise_storages())
def test_batching(store: IPromiseStore) -> None:
    s: DSTScheduler = dst(seeds=[1], durable_promise_storage=store)[0]

    s.register_command(cmd=GreetCommand, handler=batch_greeting, max_batch=2)
    s.add(
        "batching-1",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Ging",
    )
    s.add(
        "batching-2",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Razor",
    )
    s.add(
        "batching-3",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Eta",
    )
    s.add(
        "batching-4",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Elena",
    )
    s.add(
        "batching-5",
        LOptions(durable=True, retry_policy=never()),
        greet_with_batching,
        name="Dwun",
    )
    greetings_promises = s.run()
    assert all(p.success() for p in greetings_promises)
    assert [p.result() for p in greetings_promises] == [
        "Hello Ging",
        "Hello Razor",
        "Hello Eta",
        "Hello Elena",
        "Hello Dwun",
    ]


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_pin_seed(store: IPromiseStore) -> None:
    s = dst(seeds=[1], durable_promise_storage=store)[0]
    assert s.seed == 1

    os.environ[ENV_VARIABLE_PIN_SEED] = "32"
    s = dst(seeds=[1])[0]

    assert s.seed == int(os.environ.pop(ENV_VARIABLE_PIN_SEED))


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_mock_function(store: IPromiseStore) -> None:
    s = dst(seeds=[1], durable_promise_storage=store)[0]
    s.add("mock-1", LOptions(durable=True, retry_policy=never()), only_call, n=3)
    s.add("mock-2", LOptions(durable=True, retry_policy=never()), only_invocation, n=3)
    promises = s.run()
    assert all(p.result() == 3 for p in promises)  # noqa: PLR2004
    s = dst(seeds=[1], mocks={number: mocked_number})[0]
    promises = s.run()
    assert all(p.result() == 23 for p in promises)  # noqa: PLR2004


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_dst_scheduler(store: IPromiseStore) -> None:
    for _ in range(100):
        seed = random.randint(0, 1000000)  # noqa: S311
        s = dst(seeds=[seed], durable_promise_storage=store)[0]

        s.add(
            "only-call-1", LOptions(durable=True, retry_policy=never()), only_call, n=1
        )
        s.add(
            "only-call-2", LOptions(durable=True, retry_policy=never()), only_call, n=2
        )
        s.add(
            "only-call-3", LOptions(durable=True, retry_policy=never()), only_call, n=3
        )
        s.add(
            "only-call-4", LOptions(durable=True, retry_policy=never()), only_call, n=4
        )
        s.add(
            "only-call-5", LOptions(durable=True, retry_policy=never()), only_call, n=5
        )

        promises = s.run()
        values = _promise_result(promises=promises)
        assert values == [
            1,
            2,
            3,
            4,
            5,
        ], f"Test fails when seed it {seed}"


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_dst_determinitic(store: IPromiseStore) -> None:
    seed = random.randint(1, 100)  # noqa: S311
    s = dst(seeds=[seed], durable_promise_storage=store)[0]
    s.add(
        "dst-deterministic-1",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=1,
    )
    s.add(
        "dst-deterministic-2",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=2,
    )
    s.add(
        "dst-deterministic-3",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=3,
    )
    s.add(
        "dst-deterministic-4",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=4,
    )
    s.add(
        "dst-deterministic-5",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=5,
    )
    promises = s.run()
    assert sum(p.result() for p in promises) == 15  # noqa: PLR2004
    expected_events = s.get_events()

    same_seed_s = dst(seeds=[seed])[0]
    same_seed_s.add(
        "dst-deterministic-1",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=1,
    )
    same_seed_s.add(
        "dst-deterministic-2",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=2,
    )
    same_seed_s.add(
        "dst-deterministic-3",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=3,
    )
    same_seed_s.add(
        "dst-deterministic-4",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=4,
    )
    same_seed_s.add(
        "dst-deterministic-5",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=5,
    )
    promises = same_seed_s.run()
    assert sum(p.result() for p in promises) == 15  # noqa: PLR2004
    assert expected_events == same_seed_s.get_events()

    different_seed_s = dst(seeds=[seed + 10])[0]
    different_seed_s.add(
        "dst-deterministic-1",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=1,
    )
    different_seed_s.add(
        "dst-deterministic-2",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=2,
    )
    different_seed_s.add(
        "dst-deterministic-3",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=3,
    )
    different_seed_s.add(
        "dst-deterministic-4",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=4,
    )
    different_seed_s.add(
        "dst-deterministic-5",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=5,
    )
    promises = different_seed_s.run()
    assert sum(p.result() for p in promises) == 15  # noqa: PLR2004
    assert expected_events != different_seed_s.get_events()


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_failing_asserting(store: IPromiseStore) -> None:
    s = dst(seeds=[1], durable_promise_storage=store)[0]
    s.add(
        "failing-asserting-1",
        LOptions(durable=True, retry_policy=never()),
        failing_asserting,
    )
    p = s.run()
    with pytest.raises(AssertionError):
        p[0].result()


@pytest.mark.dst
@pytest.mark.parametrize("scheduler", resonate.testing.dst([range(10)]))
def test_dst_framework(scheduler: DSTScheduler) -> None:
    scheduler.add(
        "dst-framework-1", LOptions(durable=True, retry_policy=never()), only_call, n=1
    )
    scheduler.add(
        "dst-framework-2", LOptions(durable=True, retry_policy=never()), only_call, n=2
    )
    scheduler.add(
        "dst-framework-3", LOptions(durable=True, retry_policy=never()), only_call, n=3
    )
    scheduler.add(
        "dst-framework-4", LOptions(durable=True, retry_policy=never()), only_call, n=4
    )
    scheduler.add(
        "dst-framework-5", LOptions(durable=True, retry_policy=never()), only_call, n=5
    )
    promises = scheduler.run()
    assert sum(p.result() for p in promises) == 15  # noqa: PLR2004


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_failure(store: IPromiseStore) -> None:
    scheduler = dst(
        seeds=[1],
        max_failures=3,
        failure_chance=100,
        durable_promise_storage=store,
    )[0]
    scheduler.add(
        "failure-1", LOptions(durable=True, retry_policy=never()), only_call, n=1
    )
    p = scheduler.run()
    assert p[0].done()
    assert p[0].result() == 1
    assert scheduler.tick == 6  # noqa: PLR2004
    assert scheduler.current_failures == 3  # noqa: PLR2004

    scheduler = dst(seeds=[1], max_failures=2, failure_chance=0)[0]
    scheduler.add(
        "failure-1", LOptions(durable=True, retry_policy=never()), only_call, n=1
    )
    p = scheduler.run()
    assert p[0].done()
    assert p[0].result() == 1
    assert scheduler.tick == 3  # noqa: PLR2004
    assert scheduler.current_failures == 0


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_sequential_retry(store: IPromiseStore) -> None:
    seq_scheduler = dst(seeds=[1], mode="sequential", durable_promise_storage=store)[0]
    seq_scheduler.add(
        "execution-seq-with-retry",
        LOptions(durable=True, retry_policy=never()),
        only_call,
        n=1,
    )
    promises = seq_scheduler.run()
    assert [p.result() for p in promises] == [1]
    assert seq_scheduler.get_events() == [
        PromiseCreated(
            promise_id="execution-seq-with-retry", parent_promise_id=None, tick=0
        ),
        ExecutionInvoked(
            promise_id="execution-seq-with-retry",
            parent_promise_id=None,
            tick=0,
            fn_name="only_call",
            args=(),
            kwargs={"n": 1},
        ),
        PromiseCreated(
            promise_id="execution-seq-with-retry.1",
            parent_promise_id="execution-seq-with-retry",
            tick=1,
        ),
        ExecutionInvoked(
            promise_id="execution-seq-with-retry.1",
            parent_promise_id="execution-seq-with-retry",
            tick=1,
            fn_name="number",
            args=(),
            kwargs={"n": 1},
        ),
        ExecutionAwaited(
            promise_id="execution-seq-with-retry",
            parent_promise_id=None,
            tick=1,
        ),
        ExecutionTerminated(
            promise_id="execution-seq-with-retry.1",
            parent_promise_id="execution-seq-with-retry",
            tick=2,
        ),
        PromiseCompleted(
            promise_id="execution-seq-with-retry.1",
            parent_promise_id="execution-seq-with-retry",
            tick=2,
            value=Ok(1),
        ),
        ExecutionResumed(
            promise_id="execution-seq-with-retry",
            parent_promise_id=None,
            tick=3,
        ),
        ExecutionTerminated(
            promise_id="execution-seq-with-retry",
            parent_promise_id=None,
            tick=3,
        ),
        PromiseCompleted(
            promise_id="execution-seq-with-retry",
            parent_promise_id=None,
            tick=3,
            value=Ok(1),
        ),
    ]

    promises = seq_scheduler.run()
    assert [p.result() for p in promises] == [1]
    assert seq_scheduler.get_events() == [
        PromiseCreated(
            promise_id="execution-seq-with-retry", parent_promise_id=None, tick=0
        ),
        PromiseCompleted(
            promise_id="execution-seq-with-retry",
            parent_promise_id=None,
            tick=0,
            value=Ok(1),
        ),
    ]


@pytest.mark.dst
@pytest.mark.parametrize("store", _promise_storages())
def test_sequential(store: IPromiseStore) -> None:
    seq_scheduler = dst(seeds=[1], mode="sequential", durable_promise_storage=store)[0]
    seq_scheduler.add(
        "execution-seq-1", LOptions(durable=True, retry_policy=never()), only_call, n=1
    )
    seq_scheduler.add(
        "execution-seq-2", LOptions(durable=True, retry_policy=never()), only_call, n=2
    )
    seq_scheduler.add(
        "execution-seq-3", LOptions(durable=True, retry_policy=never()), only_call, n=3
    )
    seq_scheduler.add(
        "execution-seq-4", LOptions(durable=True, retry_policy=never()), only_call, n=4
    )
    seq_scheduler.add(
        "execution-seq-5", LOptions(durable=True, retry_policy=never()), only_call, n=5
    )
    promises = seq_scheduler.run()
    assert [p.result() for p in promises] == [1, 2, 3, 4, 5]

    con_scheduler = dst(seeds=[1], max_failures=2, durable_promise_storage=store)[0]
    con_scheduler.add(
        "execution-con-1", LOptions(durable=True, retry_policy=never()), only_call, n=1
    )
    con_scheduler.add(
        "execution-con-2", LOptions(durable=True, retry_policy=never()), only_call, n=2
    )
    con_scheduler.add(
        "execution-con-3", LOptions(durable=True, retry_policy=never()), only_call, n=3
    )
    con_scheduler.add(
        "execution-con-4", LOptions(durable=True, retry_policy=never()), only_call, n=4
    )
    con_scheduler.add(
        "execution-con-5", LOptions(durable=True, retry_policy=never()), only_call, n=5
    )
    promises = con_scheduler.run()
    assert [p.result() for p in promises] == [1, 2, 3, 4, 5]
    assert len(con_scheduler.get_events()) == len(seq_scheduler.get_events())


@pytest.mark.parametrize("store", _promise_storages())
def test_dump_events(store: IPromiseStore) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        log_file_path = Path(temp_dir) / "cool_log_%s.txt"
        s = dst(
            seeds=[1], log_file=log_file_path.as_posix(), durable_promise_storage=store
        )[0]
        formatted_file_path = Path(log_file_path.as_posix() % (s.seed))
        assert not formatted_file_path.exists()
        s.add(
            "dump-events-1",
            LOptions(durable=True, retry_policy=never()),
            only_call,
            n=1,
        )
        s.add(
            "dump-events-2",
            LOptions(durable=True, retry_policy=never()),
            only_call,
            n=2,
        )
        s.add(
            "dump-events-3",
            LOptions(durable=True, retry_policy=never()),
            only_call,
            n=3,
        )
        s.add(
            "dump-events-4",
            LOptions(durable=True, retry_policy=never()),
            only_call,
            n=4,
        )
        s.add(
            "dump-events-5",
            LOptions(durable=True, retry_policy=never()),
            only_call,
            n=5,
        )
        s.run()

        assert formatted_file_path.exists()
        assert formatted_file_path.read_text() == "".join(
            f"{e}\n" for e in s.get_events()
        )


def _probe_function(deps: Dependencies, tick: int) -> int:  # noqa: ARG001
    return now()


@pytest.mark.parametrize("store", _promise_storages())
def test_probe(store: IPromiseStore) -> None:
    s = dst(seeds=[1], probe=_probe_function, durable_promise_storage=store)[0]
    s.add("probe-1", LOptions(durable=True, retry_policy=never()), only_call, n=1)
    s.add("probe-2", LOptions(durable=True, retry_policy=never()), only_call, n=2)
    s.add("probe-3", LOptions(durable=True, retry_policy=never()), only_call, n=3)
    s.add("probe-4", LOptions(durable=True, retry_policy=never()), only_call, n=4)
    s.add("probe-5", LOptions(durable=True, retry_policy=never()), only_call, n=5)
    s.run()
    assert len(s.probe_results) > 0
    assert s.probe_results[-1] <= now()


def baz(ctx: Context) -> None: ...  # noqa: ARG001, RUF100
def bar(ctx: Context) -> Generator[Yieldable, Any, str]:
    yield ctx.lfc(baz)
    return "Done"


def foo(ctx: Context) -> Generator[Yieldable, Any, list[str]]:
    p1: Promise[str] = yield ctx.lfi(bar)
    p2: Promise[str] = yield ctx.lfi(bar)

    resp: list[str] = []
    v1: str = yield p1
    resp.append(v1)

    v2: str = yield p2
    resp.append(v2)
    return resp


@pytest.mark.parametrize("store", _promise_storages())
def test_events_logic(store: IPromiseStore) -> None:
    s = dst(seeds=[1], durable_promise_storage=store)[0]
    s.add("foo-events-1", LOptions(durable=True, retry_policy=never()), foo)
    s.run()
    assert s.get_events() == [
        PromiseCreated(promise_id="foo-events-1", parent_promise_id=None, tick=0),
        ExecutionInvoked(
            promise_id="foo-events-1",
            parent_promise_id=None,
            tick=0,
            fn_name="foo",
            args=(),
            kwargs={},
        ),
        PromiseCreated(
            promise_id="foo-events-1.1", parent_promise_id="foo-events-1", tick=1
        ),
        ExecutionInvoked(
            promise_id="foo-events-1.1",
            parent_promise_id="foo-events-1",
            tick=1,
            fn_name="bar",
            args=(),
            kwargs={},
        ),
        PromiseCreated(
            promise_id="foo-events-1.1.1", parent_promise_id="foo-events-1.1", tick=2
        ),
        ExecutionInvoked(
            promise_id="foo-events-1.1.1",
            parent_promise_id="foo-events-1.1",
            tick=2,
            fn_name="baz",
            args=(),
            kwargs={},
        ),
        ExecutionAwaited(
            promise_id="foo-events-1.1", parent_promise_id="foo-events-1", tick=2
        ),
        PromiseCreated(
            promise_id="foo-events-1.2", parent_promise_id="foo-events-1", tick=3
        ),
        ExecutionInvoked(
            promise_id="foo-events-1.2",
            parent_promise_id="foo-events-1",
            tick=3,
            fn_name="bar",
            args=(),
            kwargs={},
        ),
        ExecutionAwaited(promise_id="foo-events-1", parent_promise_id=None, tick=4),
        PromiseCreated(
            promise_id="foo-events-1.2.1", parent_promise_id="foo-events-1.2", tick=5
        ),
        ExecutionInvoked(
            promise_id="foo-events-1.2.1",
            parent_promise_id="foo-events-1.2",
            tick=5,
            fn_name="baz",
            args=(),
            kwargs={},
        ),
        ExecutionAwaited(
            promise_id="foo-events-1.2", parent_promise_id="foo-events-1", tick=5
        ),
        ExecutionTerminated(
            promise_id="foo-events-1.1.1", parent_promise_id="foo-events-1.1", tick=6
        ),
        PromiseCompleted(
            promise_id="foo-events-1.1.1",
            parent_promise_id="foo-events-1.1",
            tick=6,
            value=Ok(None),
        ),
        ExecutionTerminated(
            promise_id="foo-events-1.2.1", parent_promise_id="foo-events-1.2", tick=7
        ),
        PromiseCompleted(
            promise_id="foo-events-1.2.1",
            parent_promise_id="foo-events-1.2",
            tick=7,
            value=Ok(None),
        ),
        ExecutionResumed(
            promise_id="foo-events-1.1", parent_promise_id="foo-events-1", tick=8
        ),
        ExecutionTerminated(
            promise_id="foo-events-1.1", parent_promise_id="foo-events-1", tick=8
        ),
        PromiseCompleted(
            promise_id="foo-events-1.1",
            parent_promise_id="foo-events-1",
            tick=8,
            value=Ok("Done"),
        ),
        ExecutionResumed(promise_id="foo-events-1", parent_promise_id=None, tick=9),
        ExecutionAwaited(promise_id="foo-events-1", parent_promise_id=None, tick=9),
        ExecutionResumed(
            promise_id="foo-events-1.2", parent_promise_id="foo-events-1", tick=10
        ),
        ExecutionTerminated(
            promise_id="foo-events-1.2", parent_promise_id="foo-events-1", tick=10
        ),
        PromiseCompleted(
            promise_id="foo-events-1.2",
            parent_promise_id="foo-events-1",
            tick=10,
            value=Ok("Done"),
        ),
        ExecutionResumed(promise_id="foo-events-1", parent_promise_id=None, tick=11),
        ExecutionTerminated(promise_id="foo-events-1", parent_promise_id=None, tick=11),
        PromiseCompleted(
            promise_id="foo-events-1",
            parent_promise_id=None,
            tick=11,
            value=Ok(["Done", "Done"]),
        ),
    ]



from __future__ import annotations

from resonate.queue import DelayQueue


def test_delay_queue() -> None:
    queue = DelayQueue[int](caller_event=None)
    queue.put_nowait(item=3, delay=0.002)
    assert queue.dequeue() == 3  # noqa: PLR2004
    assert queue.items_in_delay() == 0
    queue.put_nowait(item=10, delay=0.01)
    queue.put_nowait(item=3, delay=0.002)
    assert queue.dequeue() == 3  # noqa: PLR2004
    assert queue.items_in_delay() == 1
    assert queue.dequeue() == 10  # noqa: PLR2004
    assert queue.items_in_delay() == 0



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from resonate.record import (
        CallbackRecord,
        DurablePromiseRecord,
        Invoke,
        Resume,
        TaskRecord,
    )
    from resonate.typing import Data, Headers, IdempotencyKey, State, Tags


class ITaskStore(ABC):
    @abstractmethod
    def claim_task(
        self, *, task_id: str, counter: int, pid: str, ttl: int
    ) -> Invoke | Resume: ...

    @abstractmethod
    def complete_task(self, *, task_id: str, counter: int) -> None: ...

    @abstractmethod
    def heartbeat_tasks(self, *, pid: str) -> int: ...

    @abstractmethod
    def create_callback(
        self,
        *,
        promise_id: str,
        root_promise_id: str,
        timeout: int,
        recv: str | dict[str, Any],
    ) -> tuple[DurablePromiseRecord, CallbackRecord | None]: ...

    @abstractmethod
    def create_with_task(  # noqa: PLR0913
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
        pid: str,
        ttl: int,
        recv: str | dict[str, Any],
    ) -> tuple[DurablePromiseRecord, TaskRecord | None]: ...

    @abstractmethod
    def create_with_callback(  # noqa: PLR0913
        self,
        *,
        promise_id: str,
        ikey: str | None,
        strict: bool,
        timeout: int,
        headers: dict[str, str] | None,
        data: str | None,
        tags: dict[str, str] | None,
        root_promise_id: str,
        recv: str | dict[str, Any],
    ) -> tuple[DurablePromiseRecord, CallbackRecord | None]: ...


class IPromiseStore(ABC):
    @abstractmethod
    def create(  # noqa: PLR0913
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
    ) -> DurablePromiseRecord: ...

    @abstractmethod
    def cancel(
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord: ...

    @abstractmethod
    def resolve(
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord: ...

    @abstractmethod
    def reject(
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord: ...

    @abstractmethod
    def get(self, *, promise_id: str) -> DurablePromiseRecord: ...

    @abstractmethod
    def search(
        self, *, promise_id: str, state: State, tags: Tags, limit: int | None = None
    ) -> list[DurablePromiseRecord]: ...



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, final


@final
@dataclass(frozen=True)
class Dependencies:
    deps: dict[str, Any] = field(default_factory=dict)

    def set(self, key: str, obj: Any) -> None:  # noqa: ANN401
        assert not self.exists(key), f"There's already a dependency under name `{key}`"
        self.deps[key] = obj

    def get(self, key: str) -> Any:  # noqa: ANN401
        return self.deps[key]

    def remove(self, key: str) -> None:
        assert self.exists(key), f"There's no depedency under name `{key}`"
        self.deps.pop(key)

    def exists(self, key: str) -> bool:
        try:
            self.get(key)
        except KeyError:
            return False
        return True



from __future__ import annotations

import asyncio
from inspect import iscoroutinefunction, isfunction
from typing import TYPE_CHECKING, Generic, TypeVar, final

from typing_extensions import ParamSpec

from resonate.result import Err, Ok, Result

if TYPE_CHECKING:
    from resonate.context import Context
    from resonate.typing import DurableAsyncFn, DurableFn, DurableSyncFn

T = TypeVar("T")
P = ParamSpec("P")


@final
class FnWrapper(Generic[T]):
    def __init__(
        self,
        ctx: Context,
        fn: DurableSyncFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.fn = fn
        self.ctx = ctx
        self.args = args
        self.kwargs = kwargs

    def run(self) -> Result[T, Exception]:
        result: Result[T, Exception]
        try:
            result = Ok(self.fn(self.ctx, *self.args, **self.kwargs))
        except Exception as e:  # noqa: BLE001
            result = Err(e)

        return result


@final
class AsyncFnWrapper(Generic[T]):
    def __init__(
        self,
        ctx: Context,
        fn: DurableAsyncFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.fn = fn
        self.ctx = ctx
        self.args = args
        self.kwargs = kwargs

    async def run(self) -> Result[T, Exception]:
        result: Result[T, Exception]
        try:
            result = Ok(asyncio.run(self.fn(self.ctx, *self.args, **self.kwargs)))
        except Exception as e:  # noqa: BLE001
            result = Err(e)
        return result


def wrap_fn(
    ctx: Context,
    fn: DurableFn[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> FnWrapper[T] | AsyncFnWrapper[T]:
    cmd: FnWrapper[T] | AsyncFnWrapper[T]
    if isfunction(fn):
        cmd = FnWrapper(ctx, fn, *args, **kwargs)
    else:
        assert iscoroutinefunction(fn)
        cmd = AsyncFnWrapper(ctx, fn, *args, **kwargs)
    return cmd



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, TypeVar

from typing_extensions import assert_never

if TYPE_CHECKING:
    from resonate.dataclasses import ResonateCoro
    from resonate.promise import Promise
    from resonate.typing import AwaitingFor

V = TypeVar("V")
K = TypeVar("K")


class Awaiting:
    def __init__(self) -> None:
        self._local: dict[Promise[Any], list[ResonateCoro[Any]]] = {}
        self._remote: dict[Promise[Any], list[ResonateCoro[Any]]] = {}

    def is_empty(self, awaiting_for: AwaitingFor) -> bool:
        if awaiting_for == "local":
            return len(self._local) == 0
        if awaiting_for == "remote":
            return len(self._remote) == 0
        assert_never(awaiting_for)

    def clear(self) -> None:
        self._local.clear()
        self._remote.clear()

    def waiting_for(
        self, key: Promise[Any], awaiting_for: AwaitingFor | None = None
    ) -> bool:
        if awaiting_for is None:
            return key in self._local or key in self._remote
        if awaiting_for == "local":
            return key in self._local
        if awaiting_for == "remote":
            return key in self._remote

        assert_never(awaiting_for)

    def append(
        self,
        key: Promise[Any],
        value: ResonateCoro[Any] | None,
        awaiting_for: AwaitingFor,
    ) -> None:
        awaiting: dict[Promise[Any], list[ResonateCoro[Any]]]
        if awaiting_for == "local":
            awaiting = self._local
        elif awaiting_for == "remote":
            awaiting = self._remote
        else:
            assert_never(awaiting_for)
        if value is not None:
            awaiting.setdefault(key, []).append(value)
        else:
            assert key not in awaiting
            awaiting[key] = []

    def get(
        self, key: Promise[Any], awaiting_for: AwaitingFor
    ) -> list[ResonateCoro[Any]] | None:
        awaiting: dict[Promise[Any], list[ResonateCoro[Any]]]
        if awaiting_for == "local":
            awaiting = self._local
        elif awaiting_for == "remote":
            awaiting = self._remote
        else:
            assert_never(awaiting_for)
        return awaiting.get(key)

    def pop(
        self, key: Promise[Any], awaiting_for: AwaitingFor
    ) -> list[ResonateCoro[Any]]:
        awaiting: dict[Promise[Any], list[ResonateCoro[Any]]]
        if awaiting_for == "local":
            awaiting = self._local
        elif awaiting_for == "remote":
            awaiting = self._remote
        else:
            assert_never(awaiting_for)
        return awaiting.pop(key)


class EphemeralMemo(Generic[K, V]):
    def __init__(self) -> None:
        self._memo: dict[K, V] = {}

    def add(self, key: K, value: V) -> None:
        assert key not in self._memo, f"There's already a value for key={key}"
        self._memo[key] = value

    def get(self, key: K) -> V | None:
        return self._memo.get(key)

    def pop(self, key: K) -> None:
        assert key in self._memo, f"There's not value for key={key}"
        self._memo.pop(key)

    def has(self, key: K) -> bool:
        return key in self._memo

    def clear(self) -> None:
        self._memo.clear()

    def is_empty(self) -> bool:
        return len(self._memo) == 0


class DoubleDict(Generic[K, V]):
    def __init__(self, data: dict[K, V] | None = None) -> None:
        self.data = data if data is not None else {}
        self._index = self._data_to_index()

    def _data_to_index(self) -> dict[V, K]:
        index = {}
        for k, v in self.data.items():
            assert v not in index, "Value cannot be duplicated."
            index[v] = k
        return index

    def add(self, key: K, value: V) -> None:
        assert key not in self.data
        assert value not in self._index
        self.data[key] = value
        self._index[value] = key

    def pop(self, key: K) -> V:
        assert key in self.data
        assert self.data[key] in self._index
        value = self.data.pop(key)
        self._index.pop(value)
        return value

    def get(self, key: K) -> V | None:
        return self.data.get(key)

    def get_from_value(self, value: V) -> K | None:
        return self._index.get(value)



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypedDict, final

from typing_extensions import Self

from resonate.encoders import IEncoder

if TYPE_CHECKING:
    from resonate.encoders import IEncoder
    from resonate.typing import Data, Headers, IdempotencyKey, State, Tags


class _InvokeInfo(TypedDict):
    func_name: str
    args: list[Any]
    kwargs: dict[str, Any]


class Decodable(ABC):
    @classmethod
    @abstractmethod
    def decode(cls, data: dict[str, Any], encoder: IEncoder[str, str]) -> Self: ...


@final
@dataclass(frozen=True)
class Param:
    data: Data
    headers: Headers


@final
@dataclass(frozen=True)
class Value:
    data: Data
    headers: Headers


@final
@dataclass(frozen=True)
class CallbackRecord(Decodable):
    callback_id: str
    promise_id: str
    timeout: int
    created_on: int

    @classmethod
    def decode(cls, data: dict[str, Any], encoder: IEncoder[str, str]) -> Self:
        _ = encoder
        return cls(
            callback_id=data["id"],
            promise_id=data["promiseId"],
            timeout=data["timeout"],
            created_on=data["createdOn"],
        )


@final
@dataclass(frozen=True)
class DurablePromiseRecord(Decodable):
    state: State
    promise_id: str
    timeout: int
    param: Param
    value: Value
    created_on: int
    completed_on: int | None
    idempotency_key_for_create: IdempotencyKey
    idempotency_key_for_complete: IdempotencyKey
    tags: Tags

    def invoke_info(self) -> _InvokeInfo:
        assert self.param.data is not None
        data_dict = json.loads(self.param.data)
        return {
            "func_name": data_dict["func"],
            "args": data_dict["args"],
            "kwargs": data_dict["kwargs"],
        }

    def is_completed(self) -> bool:
        return not self.is_pending()

    def is_timeout(self) -> bool:
        return self.state == "REJECTED_TIMEDOUT"

    def is_canceled(self) -> bool:
        return self.state == "REJECTED_CANCELED"

    def is_rejected(self) -> bool:
        return self.state == "REJECTED"

    def is_resolved(self) -> bool:
        return self.state == "RESOLVED"

    def is_pending(self) -> bool:
        return self.state == "PENDING"

    @classmethod
    def decode(cls, data: dict[str, Any], encoder: IEncoder[str, str]) -> Self:
        if data["param"]:
            param = Param(
                data=encoder.decode(data["param"]["data"]),
                headers=data["param"].get("headers"),
            )
        else:
            param = Param(data=None, headers=None)

        if data["value"]:
            value = Value(
                data=encoder.decode(data["value"]["data"]),
                headers=data["value"].get("headers"),
            )
        else:
            value = Value(data=None, headers=None)

        return cls(
            promise_id=data["id"],
            state=data["state"],
            param=param,
            value=value,
            timeout=data["timeout"],
            tags=data.get("tags"),
            created_on=data["createdOn"],
            completed_on=data.get("completedOn"),
            idempotency_key_for_complete=data.get("idempotencyKeyForComplete"),
            idempotency_key_for_create=data.get("idempotencyKeyForCreate"),
        )


@final
@dataclass(frozen=True)
class Invoke(Decodable):
    root_promise_store: DurablePromiseRecord

    @classmethod
    def decode(cls, data: dict[str, Any], encoder: IEncoder[str, str]) -> Self:
        return cls(
            root_promise_store=DurablePromiseRecord.decode(data, encoder=encoder),
        )


@final
@dataclass(frozen=True)
class Resume(Decodable):
    root_promise_store: DurablePromiseRecord
    leaf_promise_store: DurablePromiseRecord

    @classmethod
    def decode(cls, data: dict[str, Any], encoder: IEncoder[str, str]) -> Self:
        return cls(
            root_promise_store=DurablePromiseRecord.decode(
                data["root"]["data"], encoder=encoder
            ),
            leaf_promise_store=DurablePromiseRecord.decode(
                data["leaf"]["data"], encoder=encoder
            ),
        )


@final
@dataclass(frozen=True)
class TaskRecord(Decodable):
    task_id: str
    counter: int

    @classmethod
    def decode(cls, data: dict[str, Any], encoder: IEncoder[str, str]) -> Self:
        _ = encoder
        return cls(task_id=data["id"], counter=data["counter"])



from __future__ import annotations

import base64
import importlib
import json
from abc import ABC, abstractmethod
from typing import Any, Generic, final

from typing_extensions import TypeVar

In = TypeVar("In")
Out = TypeVar("Out")


class IEncoder(Generic[In, Out], ABC):
    @abstractmethod
    def encode(self, data: In) -> Out: ...

    @abstractmethod
    def decode(self, data: Out) -> In: ...


@final
class Base64Encoder(IEncoder[str, str]):
    def encode(self, data: str) -> str:
        return base64.b64encode(data.encode()).decode()

    def decode(self, data: str) -> str:
        return base64.b64decode(data).decode()


# Helper method to make Errors serializable
def _default(data: Any) -> Any:  # noqa: ANN401
    if isinstance(data, Exception):
        return {
            "__type": _classname(data),
            "attributes": data.__dict__,
        }

    return data


# Helper method to deserialize errors
def _object_hook(data: dict[str, Any]) -> Any:  # noqa: ANN401
    if "__type" in data:
        error_cls = _import_class_from_qualified_name(data["__type"])
        return error_cls(**data["attributes"])

    return data


@final
class JsonEncoder(IEncoder[Any, str]):
    def encode(self, data: Any) -> str:  # noqa: ANN401
        return json.dumps(data, default=_default)

    def decode(self, data: str) -> Any:  # noqa: ANN401
        return json.loads(data, object_hook=_object_hook)


def _classname(obj: object) -> str:
    cls = type(obj)
    module = cls.__module__
    name = cls.__qualname__
    if module is not None and module != "__builtin__":
        name = module + "." + name
    return name


def _import_class_from_qualified_name(qualified_name: str) -> Any:  # noqa: ANN401
    module_name, class_name = qualified_name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, final

import requests

from resonate.encoders import Base64Encoder, IEncoder
from resonate.errors import ResonateError
from resonate.record import (
    CallbackRecord,
    DurablePromiseRecord,
    Invoke,
    Resume,
    TaskRecord,
)
from resonate.storage.traits import IPromiseStore, ITaskStore

if TYPE_CHECKING:
    from resonate.typing import Data, Headers, IdempotencyKey, State, Tags


@final
class RemoteServer(IPromiseStore, ITaskStore):
    def __init__(
        self,
        url: str,
        request_timeout: int = 10,
        encoder: IEncoder[str, str] | None = None,
    ) -> None:
        self.url = url
        self._request_timeout = request_timeout
        self._encoder = encoder or Base64Encoder()

    def _initialize_headers(
        self, *, strict: bool, ikey: IdempotencyKey
    ) -> dict[str, str]:
        request_headers: dict[str, str] = {"Strict": str(strict)}
        if ikey is not None:
            request_headers["Idempotency-Key"] = ikey

        return request_headers

    def create_with_task(  # noqa: PLR0913
        self,
        *,
        promise_id: str,
        ikey: str | None,
        strict: bool,
        headers: dict[str, str] | None,
        data: str | None,
        timeout: int,
        tags: dict[str, str] | None,
        pid: str,
        ttl: int,
        recv: str | dict[str, Any],
    ) -> tuple[DurablePromiseRecord, TaskRecord | None]:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)
        response = requests.post(
            url=f"{self.url}/promises/task",
            headers=request_headers,
            json={
                "promise": {
                    "id": promise_id,
                    "param": {
                        "headers": headers,
                        "data": self._encode_data(data),
                    },
                    "timeout": timeout,
                    "tags": tags,
                },
                "task": {
                    "processId": pid,
                    "ttl": ttl,
                    "recv": recv,
                },
            },
            timeout=self._request_timeout,
        )
        _ensure_success(response)
        data_json = response.json()

        task_data = data_json["task"]
        durable_promise = DurablePromiseRecord.decode(
            data_json["promise"], encoder=self._encoder
        )
        if task_data is None:
            return durable_promise, None
        return durable_promise, TaskRecord.decode(task_data, encoder=self._encoder)

    def claim_task(
        self, *, task_id: str, counter: int, pid: str, ttl: int
    ) -> Invoke | Resume:
        response = requests.post(
            url=f"{self.url}/tasks/claim",
            headers={},
            json={
                "id": task_id,
                "counter": counter,
                "processId": pid,
                "ttl": ttl,
            },
            timeout=self._request_timeout,
        )
        _ensure_success(response)

        data = response.json()
        kind: Literal["resume", "invoke"] = data["type"]
        assert kind in ("resume", "invoke")

        promises: dict[str, Any] = data["promises"]
        if promises.get("leaf") is None:
            return Invoke.decode(data=promises["root"]["data"], encoder=self._encoder)

        return Resume.decode(data=promises, encoder=self._encoder)

    def create_with_callback(  # noqa: PLR0913
        self,
        *,
        promise_id: str,
        ikey: str | None,
        strict: bool,
        timeout: int,
        headers: dict[str, str] | None,
        data: str | None,
        tags: dict[str, str] | None,
        root_promise_id: str,
        recv: str | dict[str, Any],
    ) -> tuple[DurablePromiseRecord, CallbackRecord | None]:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        response = requests.post(
            url=f"{self.url}/promises/callback",
            headers=request_headers,
            json={
                "promise": {
                    "id": promise_id,
                    "timeout": timeout,
                    "param": {"headers": headers, "data": self._encode_data(data)},
                    "tags": tags,
                },
                "callback": {
                    "rootPromiseId": root_promise_id,
                    "timeout": timeout,
                    "recv": recv,
                },
            },
            timeout=self._request_timeout,
        )
        _ensure_success(response)
        data_json = response.json()

        callback_data = data_json["callback"]
        durable_promise = DurablePromiseRecord.decode(
            data_json["promise"], encoder=self._encoder
        )
        if callback_data is None:
            return durable_promise, None
        return durable_promise, CallbackRecord.decode(
            callback_data, encoder=self._encoder
        )

    def complete_task(self, *, task_id: str, counter: int) -> None:
        response = requests.post(
            url=f"{self.url}/tasks/complete",
            headers={},
            json={
                "id": task_id,
                "counter": counter,
            },
            timeout=self._request_timeout,
        )
        _ensure_success(response)

    def heartbeat_tasks(self, *, pid: str) -> int:
        response = requests.post(
            url=f"{self.url}/tasks/heartbeat",
            headers={},
            json={"processId": pid},
            timeout=self._request_timeout,
        )
        _ensure_success(response)
        return response.json()["tasksAffected"]

    def create_callback(
        self,
        *,
        promise_id: str,
        root_promise_id: str,
        timeout: int,
        recv: str | dict[str, Any],
    ) -> tuple[DurablePromiseRecord, CallbackRecord | None]:
        response = requests.post(
            url=f"{self.url}/callbacks",
            json={
                "promiseId": promise_id,
                "rootPromiseId": root_promise_id,
                "timeout": timeout,
                "recv": recv,
            },
            timeout=self._request_timeout,
        )

        _ensure_success(response)
        data = response.json()

        callback_data = data["callback"]
        durable_promise = DurablePromiseRecord.decode(
            data["promise"], encoder=self._encoder
        )
        if callback_data is None:
            return durable_promise, None
        return durable_promise, CallbackRecord.decode(
            callback_data, encoder=self._encoder
        )

    def create(  # noqa: PLR0913
        self,
        *,
        promise_id: str,
        ikey: str | None,
        strict: bool,
        headers: dict[str, str] | None,
        data: str | None,
        timeout: int,
        tags: dict[str, str] | None,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        response = requests.post(
            url=f"{self.url}/promises",
            headers=request_headers,
            json={
                "id": promise_id,
                "param": {
                    "headers": headers,
                    "data": self._encode_data(data),
                },
                "timeout": timeout,
                "tags": tags,
            },
            timeout=self._request_timeout,
        )

        _ensure_success(response)

        return DurablePromiseRecord.decode(data=response.json(), encoder=self._encoder)

    def cancel(
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)
        response = requests.patch(
            url=f"{self.url}/promises/{promise_id}",
            headers=request_headers,
            json={
                "state": "REJECTED_CANCELED",
                "value": {"headers": headers, "data": self._encode_data(data)},
            },
            timeout=self._request_timeout,
        )
        _ensure_success(response)
        return DurablePromiseRecord.decode(data=response.json(), encoder=self._encoder)

    def resolve(
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        response = requests.patch(
            f"{self.url}/promises/{promise_id}",
            headers=request_headers,
            json={
                "state": "RESOLVED",
                "value": {"headers": headers, "data": self._encode_data(data)},
            },
            timeout=self._request_timeout,
        )
        _ensure_success(response)
        return DurablePromiseRecord.decode(data=response.json(), encoder=self._encoder)

    def reject(
        self,
        *,
        promise_id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        response = requests.patch(
            f"{self.url}/promises/{promise_id}",
            headers=request_headers,
            json={
                "state": "REJECTED",
                "value": {"headers": headers, "data": self._encode_data(data)},
            },
            timeout=self._request_timeout,
        )
        _ensure_success(response)
        return DurablePromiseRecord.decode(data=response.json(), encoder=self._encoder)

    def get(self, *, promise_id: str) -> DurablePromiseRecord:
        response = requests.get(
            f"{self.url}/promises/{promise_id}", timeout=self._request_timeout
        )
        _ensure_success(response)
        return DurablePromiseRecord.decode(data=response.json(), encoder=self._encoder)

    def search(
        self, *, promise_id: str, state: State, tags: Tags, limit: int | None = None
    ) -> list[DurablePromiseRecord]:
        raise NotImplementedError

    def _encode_data(self, data: str | None) -> str | None:
        if data is None:
            return None
        return _encode(data, self._encoder)


def _ensure_success(resp: requests.Response) -> None:
    resp.raise_for_status()


def _encode(value: str, encoder: IEncoder[str, str]) -> str:
    try:
        return encoder.encode(value)
    except Exception as e:
        msg = "Encoder error"
        raise ResonateError(msg, "STORE_ENCODER", e) from e



from __future__ import annotations

import hashlib
from typing import Any, TypeVar
from uuid import UUID

T = TypeVar("T")


def string_to_ikey(string: str) -> str:
    return UUID(bytes=hashlib.sha1(string.encode("utf-8")).digest()[:16]).hex[-4:]  # noqa: S324


def recv_url(group: str, pid: str | None) -> dict[str, Any]:
    recv: dict[str, Any] = {"type": "poll", "data": {"group": group}}
    if pid is not None:
        recv["data"]["id"] = pid
    return recv



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from resonate import utils

if TYPE_CHECKING:
    from uuid import UUID


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("hi", "af4c"),
        ("bye", "df36"),
        ("money-transfer-13112-1231", "f299"),
        ("greeting-peter", "050d"),
    ],
)
def test_string_to_ikey(string: str, expected: UUID) -> None:
    assert utils.string_to_ikey(string) == expected



from __future__ import annotations

from typing import Literal

from typing_extensions import TypeAlias

ResonateErrorCode: TypeAlias = Literal[
    "UNKNOWN",
    "FETCH",
    "CANCELED",
    "TIMEDOUT",
    "ABORT",
    "STORE",
    "STORE_UNAUTHORIZED",
    "STORE_PAYLOAD",
    "STORE_FORBIDDEN",
    "STORE_NOT_FOUND",
    "STORE_ALREADY_EXISTS",
    "STORE_INVALID_STATE",
    "STORE_ENCODER",
    "USER",
]


def _error_to_code(string: ResonateErrorCode) -> int:
    error_code_map = {
        "UNKNOWN": 0,
        "FETCH": 1,
        "CANCELED": 10,
        "TIMEDOUT": 20,
        "ABORT": 30,
        "STORE": 40,
        "STORE_UNAUTHORIZED": 41,
        "STORE_PAYLOAD": 42,
        "STORE_FORBIDDEN": 43,
        "STORE_NOT_FOUND": 44,
        "STORE_ALREADY_EXISTS": 45,
        "STORE_INVALID_STATE": 46,
        "STORE_ENCODER": 47,
        "USER": 60,
    }

    return error_code_map[string]


class ResonateError(Exception):
    def __init__(
        self,
        msg: str,
        code: ResonateErrorCode,
        cause: Exception | None = None,
        *,
        retriable: bool = False,
    ) -> None:
        super().__init__(msg)
        self.msg = msg
        self.code = code
        self.cause = cause
        self.retriable = retriable

    @classmethod
    def from_error(cls, e: Exception) -> ResonateError:
        return ResonateError("Unknown error", "UNKNOWN", e, retriable=True)



from __future__ import annotations

from typing import Any

import pytest

from resonate.encoders import Base64Encoder, JsonEncoder
from resonate.errors import ResonateError


@pytest.mark.parametrize("data", ["hi", "bye", "1"])
def test_base64_encoder(data: str) -> None:
    encoder = Base64Encoder()
    encoded = encoder.encode(data)
    assert data == encoder.decode(encoded)


@pytest.mark.parametrize("data", [1, {"value": 3}, {"number": "1", "text": 312}])
def test_json_encoder(data: Any) -> None:  # noqa: ANN401
    encoder = JsonEncoder()
    encoded = encoder.encode(data=data)
    assert data == encoder.decode(encoded)


class _CustomerError(Exception):
    def __init__(self, age: int, *, testable: bool) -> None:
        super().__init__(f"I'm {age} centuries old")
        self.age = age
        self.testable = testable


@pytest.mark.parametrize(
    "data",
    [
        ResonateError("I'm here for testing", code="ABORT", retriable=True),
        _CustomerError(age=10232, testable=False),
    ],
)
def test_error_encoder(data: Exception) -> None:
    encoder = JsonEncoder()
    encoded = encoder.encode(data=data)
    assert isinstance(encoder.decode(encoded), type(data))



from __future__ import annotations

from dataclasses import dataclass
from inspect import isgenerator
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from typing_extensions import ParamSpec, assert_never

from resonate.actions import LFI
from resonate.commands import CreateDurablePromiseReq
from resonate.promise import all_promises_are_done
from resonate.result import Err, Ok

if TYPE_CHECKING:
    from collections.abc import Generator

    from resonate.context import Context
    from resonate.promise import Promise
    from resonate.result import Result
    from resonate.typing import (
        DurableCoro,
        DurableFn,
        Yieldable,
    )

T = TypeVar("T")
P = ParamSpec("P")


class RouteInfo:
    def __init__(
        self,
        ctx: Context,
        promise: Promise[Any],
        retry_attempt: int,
    ) -> None:
        assert isinstance(promise.action, LFI)
        assert isinstance(promise.action.exec_unit, FnOrCoroutine)
        self.ctx = ctx
        self.promise = promise
        self.fn_or_coroutine = promise.action.exec_unit
        self.retry_attempt = retry_attempt
        self.retry_policy = promise.action.opts.retry_policy


class ResonateCoro(Generic[T]):
    def __init__(
        self,
        route_info: RouteInfo,
        coro: Generator[Yieldable, Any, T],
    ) -> None:
        assert isgenerator(coro)
        self.route_info = route_info
        self._coro = coro
        self._coro_active = True
        self._final_value: Result[T, Exception] | None = None
        self._first_error_used = False
        self._next_child_to_yield = 0

    def _pending_children(self) -> list[Promise[Any]]:
        return self.route_info.promise.children_promises[self._next_child_to_yield :]

    def next(self) -> Yieldable:
        assert self._coro_active
        return next(self._coro)

    def send(self, v: T) -> Yieldable:
        if self._coro_active:
            try:
                return self._coro.send(v)
            except StopIteration as e:
                assert self._final_value is None, "return value can only be set once."
                self._final_value = Ok(e.value)
                self._coro_active = False

        for child in self._pending_children():
            self._next_child_to_yield += 1
            if child.done():
                continue
            return child

        assert all_promises_are_done(
            self.route_info.promise.children_promises
        ), "All children promise must have been resolved."
        assert not self._coro_active
        assert isinstance(self._final_value, Ok)
        raise StopIteration(self._final_value.unwrap())

    def throw(self, error: Exception) -> Yieldable:
        if self._coro_active:
            try:
                return self._coro.throw(error)
            except StopIteration as e:
                assert self._final_value is None, "return value can only be set once."
                self._final_value = Ok(e.value)
                self._coro_active = False

        if not self._first_error_used:
            self._final_value = Err(error)
            self._first_error_used = True

        for child in self._pending_children():
            self._next_child_to_yield += 1
            if child.done():
                continue
            return child
        assert all_promises_are_done(
            self.route_info.promise.children_promises
        ), "All children promise must have been resolved."

        assert not self._coro_active
        assert self._final_value is not None
        if isinstance(self._final_value, Ok):
            raise StopIteration(self._final_value)
        if isinstance(self._final_value, Err):
            raise self._final_value.err()
        assert_never(self._final_value)


@dataclass(frozen=True)
class Runnable(Generic[T]):
    coro: ResonateCoro[T]
    next_value: Result[Any, Exception] | None


class FnOrCoroutine:
    def __init__(
        self,
        exec_unit: DurableCoro[P, Any] | DurableFn[P, Any],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.exec_unit = exec_unit
        self.args = args
        self.kwargs = kwargs

    def to_req(
        self, promise_id: str | None, func_name: str, tags: dict[str, str]
    ) -> CreateDurablePromiseReq:
        return CreateDurablePromiseReq(
            promise_id=promise_id,
            data={
                "func": func_name,
                "args": self.args,
                "kwargs": self.kwargs,
            },
            tags=tags,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from resonate.context import Context


class Command:
    def __call__(self, ctx: Context) -> None:
        # This is not meant to be call. We are making the type system happy.
        _ = ctx
        msg = "You should never be here!"
        raise AssertionError(msg)


class CreateDurablePromiseReq(Command):
    def __init__(
        self,
        promise_id: str | None,
        data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        tags: dict[str, str] | None = None,
    ) -> None:
        self.promise_id = promise_id
        self.data = data
        self.headers = headers
        self.tags = tags


def manual_completion(promise_id: str | None) -> CreateDurablePromiseReq:
    return CreateDurablePromiseReq(promise_id=promise_id)


def remote_function(
    func_name: str,
    args: list[Any],
    *,
    target: str,
    promise_id: str | None = None,
) -> CreateDurablePromiseReq:
    return CreateDurablePromiseReq(
        promise_id=promise_id,
        data={"func": func_name, "args": args, "kwargs": {}},
        headers=None,
        tags={"resonate:invoke": f"poll://{target}"},
    )



from __future__ import annotations

import os
import sys
import uuid

import pytest

from resonate.storage.resonate_server import (
    RemoteServer,
)


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_0_callback_on_existing_promise() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    store.create(
        promise_id="0.0",
        ikey=None,
        strict=False,
        headers=None,
        data=None,
        timeout=sys.maxsize,
        tags=None,
    )
    promise_record, callback_record = store.create_callback(
        promise_id="0.0", root_promise_id="0.0", timeout=sys.maxsize, recv="default"
    )
    assert callback_record is not None
    assert promise_record.promise_id == callback_record.promise_id
    assert callback_record.timeout == sys.maxsize
    assert callback_record.promise_id == "0.0"


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_1_callback_on_non_existing_promise() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    with pytest.raises(Exception):  # noqa: B017, PT011
        store.create_callback(
            promise_id="1.1", root_promise_id="1.0", timeout=sys.maxsize, recv="default"
        )


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_2_callback_on_resolved_promise() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    store.create(
        promise_id="2.0",
        ikey=None,
        strict=False,
        headers=None,
        data=None,
        timeout=sys.maxsize,
        tags=None,
    )
    store.resolve(promise_id="2.0", ikey=None, strict=False, headers=None, data="1")
    promise_record, callback_record = store.create_callback(
        promise_id="2.0", root_promise_id="2.0", timeout=sys.maxsize, recv="default"
    )
    assert promise_record.is_completed()
    assert callback_record is None


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_3_create_durable_promise_with_task() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    pid = uuid.uuid4().hex
    ttl = 5 * 1000
    durable_promise, task_record = store.create_with_task(
        promise_id="3.0",
        ikey=None,
        strict=False,
        headers=None,
        data=None,
        timeout=sys.maxsize,
        tags=None,
        pid=pid,
        ttl=ttl,
        recv={"type": "poll", "data": {"group": "default", "id": pid}},
    )
    assert durable_promise.promise_id == "3.0"
    assert task_record is not None
    assert task_record.counter == 1
    store.heartbeat_tasks(pid=pid)
    store.complete_task(task_id=task_record.task_id, counter=task_record.counter)


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_3_create_durable_promise_with_callback() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    pid = uuid.uuid4().hex
    durable_promise, callback_record = store.create_with_callback(
        promise_id="4.0",
        ikey=None,
        strict=False,
        timeout=sys.maxsize,
        headers=None,
        data=None,
        tags=None,
        root_promise_id="4.0",
        recv={"type": "poll", "data": {"group": "default", "id": pid}},
    )
    assert callback_record is not None
    assert durable_promise.promise_id == callback_record.promise_id
    assert durable_promise.is_pending()
    assert durable_promise.timeout == callback_record.timeout


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_4_create_with_callback_dedup() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    store.create(
        promise_id="5.0",
        ikey="5.0",
        strict=False,
        headers=None,
        data=None,
        timeout=sys.maxsize,
        tags=None,
    )
    store.resolve(promise_id="5.0", ikey="5.0", strict=False, headers=None, data="1")
    promise_record, callback_record = store.create_with_callback(
        promise_id="5.0",
        root_promise_id="5.0",
        timeout=sys.maxsize,
        recv="default",
        ikey="5.0",
        strict=False,
        headers=None,
        data=None,
        tags=None,
    )
    assert promise_record.is_completed()
    assert callback_record is None



from __future__ import annotations

import time


def now() -> int:
    return time.time_ns()


def ns_to_secs(ns: float) -> float:
    return ns / 1e9


def secs_to_ns(secs: float) -> float:
    return secs * 1e9



from __future__ import annotations

import dataclasses
import os
import time
from functools import cache
from typing import TYPE_CHECKING, Any

import pytest

from resonate import scheduler
from resonate.commands import Command, CreateDurablePromiseReq
from resonate.result import Ok
from resonate.retry_policy import (
    Linear,
    constant,
    never,
)
from resonate.storage.local_store import (
    LocalStore,
    MemoryStorage,
)
from resonate.storage.resonate_server import RemoteServer

if TYPE_CHECKING:
    from collections.abc import Generator

    from resonate.context import Context
    from resonate.promise import Promise
    from resonate.storage.traits import IPromiseStore
    from resonate.typing import Yieldable


def foo(ctx: Context, name: str, sleep_time: float) -> str:  # noqa: ARG001
    time.sleep(sleep_time)
    return name


def baz(ctx: Context, name: str, sleep_time: float) -> Generator[Yieldable, Any, str]:
    p = yield ctx.lfi(foo, name=name, sleep_time=sleep_time).with_options(
        retry_policy=never()
    )
    return (yield p)


def bar(
    ctx: Context, name: str, sleep_time: float
) -> Generator[Yieldable, Any, Promise[str]]:
    p: Promise[str] = yield ctx.lfi(foo, name=name, sleep_time=sleep_time).with_options(
        retry_policy=never()
    )
    return p


@cache
def _promise_storages() -> list[IPromiseStore]:
    stores: list[IPromiseStore] = [LocalStore(MemoryStorage())]
    if os.getenv("RESONATE_STORE_URL") is not None:
        stores.append(RemoteServer(url=os.environ["RESONATE_STORE_URL"]))
    return stores


@pytest.mark.parametrize("store", _promise_storages())
def test_scheduler(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(
        durable_promise_storage=store,
    )

    s.register(baz, "baz-function", retry_policy=never())
    s.register(foo, "foo-function", retry_policy=never())

    promise: Promise[str] = s.run("baz-1", baz, name="A", sleep_time=0.02)
    s.wait_until_blocked()
    assert promise.done()
    assert promise.result() == "A"

    promise = s.run("foo-1", foo, name="B", sleep_time=0.02)
    s.wait_until_blocked()
    assert promise.done()
    assert promise.result() == "B"


@pytest.mark.parametrize("store", _promise_storages())
def test_multithreading_capabilities(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(
        processor_threads=3,
        durable_promise_storage=store,
    )
    s.register(baz, "baz")

    time_per_process: float = 0.5
    start = time.time()
    p1: Promise[str] = s.run(
        "multi-threaded-1",
        baz,
        name="A",
        sleep_time=time_per_process,
    )
    p2: Promise[str] = s.run(
        "multi-threaded-2",
        baz,
        name="B",
        sleep_time=time_per_process,
    )
    p3: Promise[str] = s.run(
        "multi-threaded-3",
        baz,
        name="C",
        sleep_time=time_per_process,
    )
    s.wait_until_blocked()
    assert p1.done()
    assert p2.done()
    assert p2.done()
    assert p1.result() == "A"
    assert p2.result() == "B"
    assert p3.result() == "C"
    assert time.time() - start <= 1


@pytest.mark.parametrize("store", _promise_storages())
def test_retry(store: IPromiseStore) -> None:
    tries = 0

    def _failing(ctx: Context, error: type[Exception]) -> None:  # noqa: ARG001, RUF100
        nonlocal tries
        tries += 1
        raise error

    def coro(
        ctx: Context, policy_info: dict[str, Any]
    ) -> Generator[Yieldable, Any, None]:
        policy = Linear(
            delay=policy_info["delay"], max_retries=policy_info["max_retries"]
        )
        yield ctx.lfc(_failing, error=RuntimeError).with_options(
            durable=False, retry_policy=policy
        )

    s = scheduler.Scheduler(durable_promise_storage=store)
    policy = Linear(delay=0, max_retries=2)

    s.register(coro, "coro-func", retry_policy=never())

    p: Promise[None] = s.run("retry-coro", coro, dataclasses.asdict(policy))
    s.wait_until_blocked()
    assert p.done()
    with pytest.raises(RuntimeError):
        assert p.result()
    assert tries == 3  # noqa: PLR2004


@pytest.mark.parametrize("store", _promise_storages())
def test_async_fibonacci(store: IPromiseStore) -> None:
    def fibonacci(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n
        p1: Promise[Any] = yield ctx.lfi(fibonacci, n - 1).with_options(
            promise_id=f"async-fibonacci-{n-1}"
        )
        p2: Promise[Any] = yield ctx.lfi(fibonacci, n - 2).with_options(
            promise_id=f"async-fibonacci-{n-2}"
        )
        v2 = yield p2
        v1 = yield p1
        return v1 + v2

    s = scheduler.Scheduler(durable_promise_storage=store)
    s.register(fibonacci, "async-fibonacci")
    n = 40
    p: Promise[int] = s.run(f"async-fibonacci-{n}", fibonacci, n)
    s.wait_until_blocked()
    assert p.result() == 102334155  # noqa: PLR2004


@pytest.mark.parametrize("store", _promise_storages())
def test_structure_concurrency(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(durable_promise_storage=store)

    child_p: Promise[str] | None = None

    def coro_that_triggers_structure_concurrency(
        ctx: Context,
    ) -> Generator[Yieldable, Any, int]:
        nonlocal child_p
        p: Promise[str] = yield ctx.lfi(foo, name="a", sleep_time=0)
        child_p = p
        return 1

    s.register(
        coro_that_triggers_structure_concurrency,
        "coro-that-triggers-structured-concurrency",
        retry_policy=never(),
    )

    p: Promise[int] = s.run(
        "structure-concurrency",
        coro_that_triggers_structure_concurrency,
    )
    s.wait_until_blocked()
    assert p.done()
    assert p.result()
    assert child_p is not None
    assert child_p.done()


@pytest.mark.parametrize("store", _promise_storages())
def test_structure_concurrency_with_failure(store: IPromiseStore) -> None:
    tries = 0

    def _failing(ctx: Context, error: type[Exception]) -> None:  # noqa: ARG001, RUF100
        nonlocal tries
        tries += 1
        raise error

    def coro_that_triggers_structure_concurrency_and_fails(
        ctx: Context,
    ) -> Generator[Yieldable, Any, int]:
        yield ctx.lfi(_failing, error=RuntimeError).with_options(
            retry_policy=constant(delay=0.03, max_retries=2), durable=False
        )
        return 1

    s = scheduler.Scheduler(durable_promise_storage=store)

    s.register(
        coro_that_triggers_structure_concurrency_and_fails,
        "coro-structured-concurrency-and-failure",
        retry_policy=never(),
    )
    p: Promise[int] = s.run(
        "structure-concurrency-with-failure",
        coro_that_triggers_structure_concurrency_and_fails,
    )
    s.wait_until_blocked()
    assert p.done()
    with pytest.raises(RuntimeError):
        p.result()

    assert tries == 3  # noqa: PLR2004


@pytest.mark.parametrize("store", _promise_storages())
def test_structure_concurrency_with_multiple_failures(store: IPromiseStore) -> None:
    def _failing(ctx: Context, error: type[Exception]) -> None:  # noqa: ARG001, RUF100
        raise error

    def coro_that_trigger_structure_concurrency_and_multiple_errors(
        ctx: Context,
    ) -> Generator[Yieldable, Any, int]:
        yield ctx.lfi(_failing, error=TypeError).with_options(
            retry_policy=never(), durable=False
        )
        yield ctx.lfi(_failing, error=RuntimeError).with_options(
            durable=False, retry_policy=never()
        )
        return 1

    s = scheduler.Scheduler(durable_promise_storage=store)

    s.register(
        coro_that_trigger_structure_concurrency_and_multiple_errors,
        "ss-and-multiple-errors",
        retry_policy=never(),
    )
    p: Promise[int] = s.run(
        "structure-concurrency-with-many-failure",
        coro_that_trigger_structure_concurrency_and_multiple_errors,
    )
    s.wait_until_blocked()
    assert p.done()
    with pytest.raises(TypeError):
        p.result()


@pytest.mark.parametrize("store", _promise_storages())
def test_deferred_invoke(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(durable_promise_storage=store)

    deferred_p: Promise[str] | None = None

    def coro_with_deferred_invoke(ctx: Context) -> Generator[Yieldable, Any, int]:
        nonlocal deferred_p
        p: Promise[str] = yield ctx.deferred(
            "deferred-invoke", foo, name="A", sleep_time=0.02
        )
        deferred_p = p

        return 1

    s.register(coro_with_deferred_invoke, "coro-with-deferred-invoked")
    s.register(foo, "foo")

    p: Promise[int] = s.run("test-deferred-invoke", coro_with_deferred_invoke)
    s.wait_until_blocked()
    assert p.done()
    assert p.result() == 1
    assert deferred_p is not None
    assert deferred_p.done()

    def_p = s.run("deferred-invoke", foo, name="B", sleep_time=120)
    assert def_p.done()
    assert def_p.result() == "A"


def _fn_sleep(_ctx: Context, wait: float, result: str) -> str:
    time.sleep(wait)
    return result


def _fn_raise_or_ret(_ctx: Context, val: Any) -> Any:  # noqa: ANN401
    if isinstance(val, Exception):
        raise val

    return val


def race_coro(
    ctx: Context, waits_results: list[tuple[float, str]]
) -> Generator[Yieldable, Any, str]:
    ps = []
    for wt, result in waits_results:
        p = yield ctx.lfi(_fn_sleep, wait=wt, result=result)
        ps.append(p)

    p_race = yield ctx.race(ps)

    res = yield p_race
    return res


def all_coro(
    ctx: Context, waits_results: list[tuple[float, str]]
) -> Generator[Yieldable, Any, list[str]]:
    ps = []
    for wt, result in waits_results:
        p = yield ctx.lfi(_fn_sleep, wait=wt, result=result)
        ps.append(p)

    p_all = yield ctx.all(ps)

    res = yield p_all
    return res


def all_settled_coro(
    ctx: Context, vals: list[Any]
) -> Generator[Yieldable, Any, list[str]]:
    ps = []
    for val in vals:
        p = yield ctx.lfi(_fn_raise_or_ret, val=val).with_options(retry_policy=never())
        ps.append(p)

    p_all_settled = yield ctx.all_settled(ps)

    res = yield p_all_settled
    return res


@pytest.mark.parametrize("store", _promise_storages())
def test_all_combinator(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(durable_promise_storage=store)
    # Test case 1
    waits_results = [(0.02, "A"), (0.03, "B"), (0.01, "C"), (0.02, "D"), (0.02, "E")]
    expected = ["A", "B", "C", "D", "E"]
    s.register(all_coro, "all-coro")
    p_all: Promise[list[str]] = s.run(
        "all-coro-0", all_coro, waits_results=waits_results
    )
    s.wait_until_blocked()
    assert p_all.done()
    assert p_all.result() == expected

    # Test case 2
    waits_results = [(0.05, "A"), (0.04, "B"), (0.03, "C"), (0.02, "D"), (0.01, "E")]
    expected = ["A", "B", "C", "D", "E"]
    p_all = s.run("all-coro-1", all_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_all.done()
    assert p_all.result() == expected

    # Test case 3
    waits_results = [(0.01, "A"), (0.01, "B"), (0.01, "C")]
    expected = ["A", "B", "C"]
    p_all = s.run("all-coro-2", all_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_all.done()
    assert p_all.result() == expected

    # Test case 4
    waits_results = [(0.1, "A")]
    expected = ["A"]
    p_all = s.run("all-coro-3", all_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_all.done()
    assert p_all.result() == expected

    # Test case - empty list
    waits_results = []
    expected = []
    p_all = s.run("all-coro-4", all_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_all.done()
    assert p_all.result() == expected


@pytest.mark.parametrize("store", _promise_storages())
def test_all_settled_combinator(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(durable_promise_storage=store)
    # Test case 1
    vals = ["A", "B", "C", "D", "E"]
    expected = ["A", "B", "C", "D", "E"]
    s.register(all_settled_coro, "all-settled-coro")
    p_all_settled: Promise[list[Any]] = s.run(
        "all-settled-coro-0", all_settled_coro, vals=vals
    )
    s.wait_until_blocked()
    assert p_all_settled.done()
    assert p_all_settled.result() == expected

    # Test case 2
    vals = ["A"]
    expected = ["A"]
    p_all_settled = s.run("all-settled-coro-1", all_settled_coro, vals=vals)
    s.wait_until_blocked()
    assert p_all_settled.done()
    assert p_all_settled.result() == expected

    # Test case 3
    vals_err = ["A", ValueError()]
    expected_err = ["A", ValueError()]
    p_all_settled = s.run("all-settled-coro-2", all_settled_coro, vals=vals_err)
    for val, exp in zip(p_all_settled.result(timeout=1), expected_err):
        if isinstance(val, Exception):
            assert isinstance(val, type(exp))
        else:
            assert val == exp

    # Test case - empty list
    vals = []
    expected = []
    p_all_settled = s.run("all-settled-coro-3", all_settled_coro, vals=vals)
    s.wait_until_blocked()
    assert p_all_settled.done()
    assert p_all_settled.result() == expected


@pytest.mark.parametrize("store", _promise_storages())
def test_race_combinator(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(durable_promise_storage=store, processor_threads=16)

    # Test case 1
    waits_results = [(0.1, "A"), (0.1, "B"), (0.01, "C"), (0.1, "D"), (0.1, "E")]
    expected = "C"
    s.register(race_coro, "race-coro", retry_policy=never())
    p_race: Promise[str] = s.run("race-coro-0", race_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_race.done()
    assert p_race.result() == expected

    # Test case 2
    waits_results = [(0.1, "A"), (0.1, "B"), (0.1, "C"), (0.1, "D"), (0.01, "E")]
    expected = "E"
    p_race = s.run("race-coro-1", race_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_race.done()
    assert p_race.result() == expected

    # # Test case 3
    waits_results = [(0.01, "A"), (0.1, "B"), (0.1, "C")]
    expected = "A"
    p_race = s.run("race-coro-2", race_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_race.done()
    assert p_race.result() == expected

    # Test case 4
    waits_results = [(0.1, "A")]
    expected = "A"
    p_race = s.run("race-coro-3", race_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_race.done()
    assert p_race.result() == expected

    # Test case 5
    waits_results = []
    p_race = s.run("race-coro-4", race_coro, waits_results=waits_results)
    s.wait_until_blocked()
    assert p_race.done()
    with pytest.raises(AssertionError):
        p_race.result()


@pytest.mark.parametrize("store", _promise_storages())
def test_coro_retry(store: IPromiseStore) -> None:
    s = scheduler.Scheduler(durable_promise_storage=store)

    tries = 0

    def _coro_to_retry(ctx: Context) -> Generator[Yieldable, Any, int]:
        nonlocal tries
        tries += 1
        yield ctx.lfc(foo, name="A", sleep_time=0)
        raise ValueError

    s.register(_coro_to_retry, "coro-to-retry", constant(delay=0, max_retries=3))
    p: Promise[int] = s.run("coro-to-retry.1", _coro_to_retry)
    s.wait_until_blocked()
    assert p.done()
    with pytest.raises(ValueError):  # noqa: PT011
        p.result()

    assert tries == 4  # noqa: PLR2004


def _raw_rfc(ctx: Context) -> Generator[Yieldable, Any, None]:
    yield ctx.rfc(
        CreateDurablePromiseReq(
            promise_id="abc",
            data={
                "func": "func",
                "args": (1, 2),
            },
            tags={"demo": "test"},
        )
    )


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_rfc_raw() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])

    s = scheduler.Scheduler(durable_promise_storage=store)
    s.register(_raw_rfc)
    p: Promise[None] = s.run("test-raw-rfc", _raw_rfc)
    s.wait_until_blocked()
    assert not p.done()

    child_promise_record = store.get(promise_id="abc")
    assert child_promise_record.promise_id == "abc"
    assert child_promise_record.state == "PENDING"
    assert child_promise_record.tags == {"demo": "test"}


@pytest.mark.parametrize("store", _promise_storages())
def test_dedup(store: IPromiseStore) -> None:
    def factorial(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return 1

        p: Promise[int] = yield ctx.lfi(factorial, n - 1).with_options(
            promise_id=f"factorial-dedup-{n-1}", retry_policy=never()
        )
        return n * (yield p)

    s = scheduler.Scheduler(durable_promise_storage=store)
    s.register(factorial, retry_policy=never())
    n = 5
    p1: Promise[int] = s.run(f"factorial-dedup-{n}", factorial, n)
    p2: Promise[int] = s.run(f"factorial-dedup-{n}", factorial, n)
    s.wait_until_blocked()
    assert p1.done()
    assert p1 == p2
    assert p2.result() == 120  # noqa: PLR2004


@pytest.mark.parametrize("store", _promise_storages())
def test_batching(store: IPromiseStore) -> None:
    @dataclasses.dataclass(frozen=True)
    class GreetCommand(Command):
        name: str

    def command_handler(ctx: Context, cmds: list[GreetCommand]) -> list[str]:  # noqa: ARG001
        return [cmd.name for cmd in cmds]

    def greet(ctx: Context, name: str) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.lfi(GreetCommand(name=name))
        v: str = yield p
        return v

    s = scheduler.Scheduler(store)
    s.register_command_handler(GreetCommand, command_handler, maxlen=10)
    s.register(greet)

    characters = [
        "Gon Freecss",
        "Killua Zoldyck",
        "Kurapika",
        "Leorio Paradinight",
        "Hisoka Morow",
        "Chrollo Lucilfer",
        "Meruem",
        "Isaac Netero",
        "Biscuit Krueger",
        "Feitan Portor",
        "Shalnark",
        "Shizuku Murasaki",
        "Nobunaga Hazama",
        "Pakunoda",
        "Phinks Magcub",
        "Uvogin",
        "Machi Komacine",
        "Kite",
        "Knuckle Bine",
        "Morel Mackernasey",
    ]
    promises: list[Promise[str]] = []
    for name in characters:
        p: Promise[str] = s.run(f"test-batching-greet-{name}", greet, name)
        promises.append(p)

    s.wait_until_blocked()

    for p, name in zip(promises, characters):
        assert p.done()
        assert p.result() == name


@pytest.mark.parametrize("store", _promise_storages())
def test_batching_with_failing_func(store: IPromiseStore) -> None:
    @dataclasses.dataclass(frozen=True)
    class ACommand(Command):
        n: int

    def command_handler(ctx: Context, cmds: list[ACommand]) -> list[str]:  # noqa: ARG001
        raise RuntimeError

    def do_something(ctx: Context, n: int) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.lfi(ACommand(n))
        v: str = yield p
        return v

    s = scheduler.Scheduler(store)

    s.register_command_handler(
        ACommand,
        command_handler,
        maxlen=10,
        retry_policy=never(),
    )
    s.register(do_something, retry_policy=never())

    promises: list[Promise[str]] = []
    for n in range(10):
        p: Promise[str] = s.run(f"do-something-{n}", do_something, n)
        promises.append(p)
    s.wait_until_blocked()

    for p in promises:
        assert p.done()
        with pytest.raises(RuntimeError):
            p.result()


@pytest.mark.parametrize("store", _promise_storages())
def test_batching_with_no_result(store: IPromiseStore) -> None:
    @dataclasses.dataclass(frozen=True)
    class ACommand(Command):
        n: int

    def command_handler(ctx: Context, cmds: list[ACommand]) -> None:  # noqa: ARG001
        _ = cmds

    def do_something(ctx: Context, n: int) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.lfi(ACommand(n))
        v: str = yield p
        return v

    s = scheduler.Scheduler(store)
    s.register_command_handler(ACommand, command_handler, maxlen=10)
    s.register(do_something, retry_policy=never())

    promises: list[Promise[str]] = []
    for n in range(10):
        p: Promise[str] = s.run(f"do-something-with-no-result-{n}", do_something, n)
        promises.append(p)
    s.wait_until_blocked()

    for p in promises:
        assert p.done()
        assert p.result() is None


@pytest.mark.parametrize("store", _promise_storages())
def test_batching_with_single_result(store: IPromiseStore) -> None:
    @dataclasses.dataclass(frozen=True)
    class ACommand(Command):
        n: int

    def command_handler(ctx: Context, cmds: list[ACommand]) -> str:  # noqa: ARG001
        _ = cmds
        return "Ok"

    def do_something(ctx: Context, n: int) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.lfi(ACommand(n))
        v: str = yield p
        return v

    s = scheduler.Scheduler(store)
    s.register_command_handler(ACommand, command_handler, maxlen=10)
    s.register(do_something, retry_policy=never())

    promises: list[Promise[str]] = []
    for n in range(10):
        p: Promise[str] = s.run(f"do-something-with-single-result-{n}", do_something, n)
        promises.append(p)

    s.wait_until_blocked()

    for p in promises:
        assert p.done()
        assert p.result() == "Ok"


@pytest.mark.parametrize("store", _promise_storages())
def test_batching_with_failing_func_and_retries(store: IPromiseStore) -> None:
    @dataclasses.dataclass(frozen=True)
    class ACommand(Command):
        n: int

    retries: int = -1

    def command_handler(ctx: Context, cmds: list[ACommand]) -> list[str]:  # noqa: ARG001
        nonlocal retries
        retries += 1
        raise RuntimeError

    def do_something(ctx: Context, n: int) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.lfi(ACommand(n))
        v: str = yield p
        return v

    s = scheduler.Scheduler(store)

    max_retries = 10
    s.register_command_handler(
        ACommand,
        command_handler,
        maxlen=10,
        retry_policy=constant(delay=0, max_retries=max_retries),
    )
    s.register(do_something, retry_policy=never())

    n = 1
    p: Promise[str] = s.run(f"test-batch-retries{n}", do_something, n)
    s.wait_until_blocked()
    assert p.done()

    with pytest.raises(RuntimeError):
        p.result()

    assert retries == max_retries


@pytest.mark.parametrize("store", _promise_storages())
def test_batching_with_element_level_exception(store: IPromiseStore) -> None:
    @dataclasses.dataclass(frozen=True)
    class ACommand(Command):
        n: int

    should_fail: bool = False

    def command_handler(ctx: Context, cmds: list[ACommand]) -> list[str | Exception]:  # noqa: ARG001
        nonlocal should_fail
        results: list[str | Exception] = []
        for _ in range(len(cmds)):
            if should_fail:
                results.append(RuntimeError("something bad happened"))
            else:
                results.append("something good happened")
            should_fail = not should_fail
        return results

    def do_something(ctx: Context, n: int) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.lfi(ACommand(n))
        v: str = yield p
        return v

    s = scheduler.Scheduler(store)

    s.register_command_handler(
        ACommand,
        command_handler,
        retry_policy=never(),
    )
    s.register(do_something, retry_policy=never())

    promises: list[Promise[str]] = []
    for i in range(10):
        promises.append(s.run(f"test-batch-element-level-failure{i}", do_something, i))  # noqa: PERF401

    s.wait_until_blocked()

    successes = 0
    failures = 0
    for p in promises:
        assert p.done()
        result = p.safe_result()
        if isinstance(result, Ok):
            successes += 1
        else:
            failures += 1

    assert successes == failures


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_remote_call_same_node() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])

    def _number_from_other_node(ctx: Context) -> int:  # noqa: ARG001
        return 1

    def _remotely(ctx: Context) -> Generator[Yieldable, Any, int]:
        n = yield ctx.rfc(_number_from_other_node).with_options(target="call-same-node")
        return n

    s = scheduler.Scheduler(store, logic_group="call-same-node")
    s.register(_remotely, retry_policy=never())
    s.register(_number_from_other_node, retry_policy=never())
    p: Promise[int] = s.run("test-remote-call-same-node", _remotely)
    s.wait_until_blocked()
    assert not p.done()
    assert p.result() == 1


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_remote_invocation_same_node() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])

    def _number_from_other_node(ctx: Context) -> int:  # noqa: ARG001
        return 1

    def _remotely(ctx: Context) -> Generator[Yieldable, Any, int]:
        p = yield ctx.rfi(_number_from_other_node).with_options(
            target="invocation-same-node"
        )
        n = yield p
        return n

    s = scheduler.Scheduler(store, logic_group="invocation-same-node")
    s.register(_remotely, retry_policy=never())
    s.register(_number_from_other_node, retry_policy=never())
    p: Promise[int] = s.run("test-remote-invocation-same-node", _remotely)
    s.wait_until_blocked()
    assert not p.done()
    assert p.result() == 1


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_remote_invocation_other_node() -> None:
    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])

    def _number_from_other_node(ctx: Context) -> int:  # noqa: ARG001
        return 1

    def _remotely(ctx: Context) -> Generator[Yieldable, Any, int]:
        p = yield ctx.rfi(_number_from_other_node).with_options(
            target="invocation-other-node"
        )
        n = yield p
        return n

    s = scheduler.Scheduler(store, logic_group="invocation-this-node")
    s.register(_remotely, retry_policy=never())
    s.register(_number_from_other_node, retry_policy=never())

    s_other = scheduler.Scheduler(store, logic_group="invocation-other-node")
    s_other.register(_remotely, retry_policy=never())
    s_other.register(_number_from_other_node, retry_policy=never())

    p: Promise[int] = s.run("test-remote-invocation-other-node", _remotely)
    s.wait_until_blocked()
    assert not p.done()
    assert p.result() == 1



from __future__ import annotations

from resonate.collections import DoubleDict


def test_double_dict() -> None:
    dd = DoubleDict[str, int]({"a": 1, "b": 2})
    assert dd.get("a") == 1
    dd.add("c", 3)
    assert dd.get("c") == 3  # noqa: PLR2004
    assert dd.get_from_value(3) == "c"
    assert dd.pop("a") == 1
    assert dd.data == {"b": 2, "c": 3}



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Callable

from typing_extensions import ParamSpec, assert_never

from resonate import random
from resonate.contants import ENV_VARIABLE_PIN_SEED
from resonate.dst.scheduler import DSTScheduler, Mode
from resonate.storage.local_store import LocalStore

if TYPE_CHECKING:
    from resonate.dependency_injection import Dependencies
    from resonate.storage.traits import IPromiseStore
    from resonate.typing import DurableCoro, DurableFn, MockFn

P = ParamSpec("P")


def dst(  # noqa: PLR0913
    seeds: list[range | int],
    mocks: dict[DurableCoro[P, Any] | DurableFn[P, Any], MockFn[Any]] | None = None,
    log_file: str | None = None,
    mode: Mode = "concurrent",
    failure_chance: float = 0,
    max_failures: int = 0,
    probe: Callable[[Dependencies, int], Any] | None = None,
    assert_always: Callable[[Dependencies, int, int], Any] | None = None,
    assert_eventually: Callable[[Dependencies, int], Any] | None = None,
    durable_promise_storage: IPromiseStore | None = None,
) -> list[DSTScheduler]:
    def _new_dst_scheduler(seed: int) -> DSTScheduler:
        return DSTScheduler(
            random=random.Random(seed),
            mocks=mocks,
            mode=mode,
            failure_chance=failure_chance,
            max_failures=max_failures,
            log_file=log_file,
            probe=probe,
            assert_always=assert_always,
            assert_eventually=assert_eventually,
            durable_promise_storage=durable_promise_storage
            if durable_promise_storage
            else LocalStore(),
        )

    schedulers: list[DSTScheduler] = []
    pin_seed = os.environ.get(ENV_VARIABLE_PIN_SEED)
    if pin_seed is not None:
        seeds = [int(pin_seed)]

    for seed in seeds:
        if isinstance(seed, range):
            schedulers.extend(_new_dst_scheduler(i) for i in seed)
        elif isinstance(seed, int):
            schedulers.append(_new_dst_scheduler(seed))
        else:
            assert_never(seed)

    return schedulers



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union, final

from typing_extensions import TypeAlias, assert_never


class Retriable(ABC):
    @abstractmethod
    def calculate_delay(self, attempt: int) -> float: ...

    @abstractmethod
    def should_retry(self, attempt: int) -> bool: ...


@final
@dataclass(frozen=True)
class Exponential(Retriable):
    """A retry policy where the delay between retries grows exponentially."""

    base_delay: float
    factor: float
    max_retries: int

    def calculate_delay(self, attempt: int) -> float:
        return self.base_delay * (self.factor**attempt)

    def should_retry(self, attempt: int) -> bool:
        return attempt <= self.max_retries


@final
@dataclass(frozen=True)
class Linear(Retriable):
    """A retry policy where the delay between retries grows linearly."""

    delay: float
    max_retries: int

    def calculate_delay(self, attempt: int) -> float:
        return self.delay * attempt

    def should_retry(self, attempt: int) -> bool:
        return attempt <= self.max_retries


@final
@dataclass(frozen=True)
class Constant(Retriable):
    """A retry policy where the delay between retries is constant."""

    delay: float
    max_retries: int

    def calculate_delay(self, attempt: int) -> float:
        _ = attempt
        return self.delay

    def should_retry(self, attempt: int) -> bool:
        return attempt <= self.max_retries


@final
@dataclass(frozen=True)
class Never:
    """A retry policy where there's no retry."""


RetryPolicy: TypeAlias = Union[Exponential, Linear, Constant, Never]


def linear(delay: float, max_retries: int) -> Linear:
    return Linear(delay=delay, max_retries=max_retries)


def exponential(base_delay: float, factor: float, max_retries: int) -> Exponential:
    return Exponential(base_delay=base_delay, factor=factor, max_retries=max_retries)


def constant(delay: float, max_retries: int) -> Constant:
    return Constant(delay=delay, max_retries=max_retries)


def never() -> Never:
    return Never()


def default_policy() -> RetryPolicy:
    return exponential(base_delay=1, factor=2, max_retries=10)


def calculate_total_possible_delay(policy: Exponential | Linear | Constant) -> float:
    total_possible_delay: float
    if isinstance(policy, Exponential):
        # Geometric series sum formula
        total_possible_delay = policy.base_delay * (
            (policy.factor * (1 - policy.factor**policy.max_retries))
            / (1 - policy.factor)
        )
    elif isinstance(policy, Linear):
        total_possible_delay = (
            policy.delay * (policy.max_retries * (policy.max_retries + 1)) / 2
        )
    elif isinstance(policy, Constant):
        total_possible_delay = policy.delay * policy.max_retries
    else:
        assert_never(policy)
    return total_possible_delay



from __future__ import annotations

import json
from threading import Thread
from typing import TYPE_CHECKING

import requests

from resonate.encoders import JsonEncoder
from resonate.record import TaskRecord

if TYPE_CHECKING:
    from resonate.scheduler import Scheduler


class LongPoller:
    def __init__(
        self,
        scheduler: Scheduler,
        url: str = "http://localhost",
    ) -> None:
        self._url = url
        self._scheduler = scheduler
        self._encoder = JsonEncoder()
        self._worket_thread = Thread(target=self._run, daemon=True)
        self._worket_thread.start()

    def _run(self) -> None:
        while True:
            poll_url = (
                f"{self._url}:8002/{self._scheduler.logic_group}/{self._scheduler.pid}"
            )
            response = requests.get(  # noqa: S113
                url=poll_url,
                headers={"Accept": "text/event-stream"},
                stream=True,
            )
            for line in response.iter_lines(chunk_size=None, decode_unicode=True):
                if not line:
                    continue
                stripped: str = line.strip()
                assert stripped.startswith("data:")
                info = json.loads(stripped[5:])
                self._scheduler.enqueue_task_record(
                    TaskRecord.decode(info["task"], encoder=self._encoder)
                )



from __future__ import annotations

from typing import final

from resonate.retry_policy import RetryPolicy, default_policy


@final
class ROptions:
    def __init__(
        self, promise_id: str | None = None, target: str | None = None
    ) -> None:
        self.promise_id = promise_id
        self.target = target


@final
class LOptions:
    def __init__(
        self,
        *,
        durable: bool = True,
        promise_id: str | None = None,
        retry_policy: RetryPolicy | None = None,
        tags: dict[str, str] | None = None,
    ) -> None:
        self.durable = durable
        self.promise_id = promise_id
        self.retry_policy = (
            retry_policy if retry_policy is not None else default_policy()
        )
        self.tags = tags or {}



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, final

from typing_extensions import ParamSpec, TypeVar, assert_never

from resonate.result import Err, Ok, Result

if TYPE_CHECKING:
    from resonate.dataclasses import Runnable
    from resonate.typing import Yieldable

T = TypeVar("T")
P = ParamSpec("P")


@final
@dataclass(frozen=True)
class FinalValue(Generic[T]):
    v: Result[T, Exception]


def iterate_coro(runnable: Runnable[T]) -> Yieldable | FinalValue[T]:
    yieldable: Yieldable
    try:
        if runnable.next_value is None:
            # `None` means to initialize the coroutine
            yieldable = runnable.coro.next()
        elif isinstance(runnable.next_value, Ok):
            # `Ok[T]` means send a value
            yieldable = runnable.coro.send(runnable.next_value.unwrap())
        elif isinstance(runnable.next_value, Err):
            # `Err[Exception]` means throw and Exception
            yieldable = runnable.coro.throw(runnable.next_value.err())
        else:
            assert_never(runnable.next_value)
    except StopIteration as e:
        return FinalValue(Ok(e.value))
    except Exception as e:  # noqa: BLE001
        return FinalValue(Err(e))
    return yieldable



from __future__ import annotations

import json
import os
import time
from random import randint
from typing import TYPE_CHECKING, Any

import pytest

from resonate.commands import manual_completion, remote_function
from resonate.promise import Promise
from resonate.retry_policy import never
from resonate.scheduler import Scheduler
from resonate.storage.resonate_server import RemoteServer

if TYPE_CHECKING:
    from collections.abc import Generator

    from resonate.context import Context
    from resonate.promise import Promise
    from resonate.typing import Yieldable


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_human_in_the_loop() -> None:
    def human_in_the_loop(ctx: Context) -> Generator[Yieldable, Any, str]:
        name: str = yield ctx.rfc(
            manual_completion("test-human-in-loop-question-to-answer-1")
        )
        age: int = yield ctx.rfc(
            manual_completion(promise_id="test-human-in-loop-question-to-answer-2")
        )
        return f"Hi {name} with age {age}"

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    s = Scheduler(store)
    s.register(human_in_the_loop, retry_policy=never())
    p: Promise[str] = s.run("test-feature-human-in-the-loop", human_in_the_loop)
    s.wait_until_blocked()
    store.resolve(
        promise_id="test-human-in-loop-question-to-answer-1",
        ikey=None,
        strict=False,
        headers=None,
        data=json.dumps("Peter"),
    )
    s.clear_blocked_flag()
    s.wait_until_blocked()
    store.resolve(
        promise_id="test-human-in-loop-question-to-answer-2",
        ikey=None,
        strict=False,
        headers=None,
        data=json.dumps(50),
    )
    s.clear_blocked_flag()
    s.wait_until_blocked()
    assert p.done()
    assert p.result() == "Hi Peter with age 50"


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_factorial_same_node() -> None:
    node_group = "test-factorial-same-node"

    def factorial(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        return n * (
            yield ctx.rfc(factorial, n - 1).with_options(
                f"factorial-same-node-{n-1}", target=node_group
            )
        )

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    s = Scheduler(store, logic_group=node_group)
    s.register(factorial)
    n = 5
    p: Promise[int] = s.run(f"factorial-same-node-{n}", factorial, n)
    s.wait_until_blocked()
    assert not p.done()
    assert p.result() == 120  # noqa: PLR2004


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_factorial_multi_node() -> None:
    def factorial_node_1(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        return n * (
            yield ctx.rfc(factorial_node_1, n - 1).with_options(
                f"factorial-multi-node-{n-1}", target="test-factorial-multi-node-2"
            )
        )

    def factorial_node_2(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        return n * (
            yield ctx.rfc(factorial_node_2, n - 1).with_options(
                f"factorial-multi-node-{n-1}", target="test-factorial-multi-node-1"
            )
        )

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    s1 = Scheduler(store, logic_group="test-factorial-multi-node-1")
    s1.register(factorial_node_1, name="factorial")
    s2 = Scheduler(store, logic_group="test-factorial-multi-node-2")
    s2.register(factorial_node_2, name="factorial")
    n = 5
    p: Promise[int] = s1.run(f"factorial-multi-node-{n}", factorial_node_1, n)
    assert p.result() == 120  # noqa: PLR2004


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_trigger_on_other_node() -> None:
    def workflow(ctx: Context) -> Generator[Yieldable, Any, str]:
        v1: int = yield ctx.rfc(
            remote_function(
                func_name="foo",
                args=[1],
                target="test-trigger-on-other-node-other",
            ),
        )
        v2: int = yield ctx.rfc(
            remote_function(
                func_name="bar",
                args=["Killua"],
                target="test-trigger-on-other-node-other",
            )
        )
        return f"{v2} is {v1}"

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    s1 = Scheduler(store, logic_group="test-trigger-on-other-node")
    s1.register(workflow, retry_policy=never())

    def _foo(ctx: Context, n: int) -> int:  # noqa: ARG001
        return n

    def _bar(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    s2 = Scheduler(store, logic_group="test-trigger-on-other-node-other")
    s2.register(_foo, retry_policy=never(), name="foo")
    s2.register(_bar, retry_policy=never(), name="bar")
    p: Promise[str] = s1.run("test-trigger-on-other-node", workflow)
    s1.wait_until_blocked()
    assert not p.done()
    assert p.result() == "Killua is 1"


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_factorial_mechanics() -> None:
    node_group = "test-factorial-mechanics"

    def factorial(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        promise_id = f"factorial-mechanics-{n-1}"
        v: int
        if n % randint(1, 10) == 0:  # noqa: S311
            v = yield ctx.rfc(factorial, n - 1).with_options(
                promise_id=promise_id, target=node_group
            )
        else:
            v = yield ctx.lfc(factorial, n - 1).with_options(
                promise_id=promise_id, retry_policy=never()
            )

        return n + v

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    schedulers: list[Scheduler] = []
    for _ in range(randint(1, 5)):  # noqa: S311
        s = Scheduler(store, logic_group=node_group)
        s.register(factorial, retry_policy=never())
        schedulers.append(s)
    n = randint(10, 30)  # noqa: S311

    p: Promise[int] = schedulers[0].run(f"factorial-mechanics-{n}", factorial, n)
    assert p.result()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_serverless_mechanics() -> None:
    node_group = "test-serverless-mechanics"
    lambda_node_group = "test-serverless-mechanics-lambda"

    def workflow(ctx: Context) -> Generator[Yieldable, Any, int]:
        v: int = yield ctx.rfc(
            remote_function(
                promise_id=None,
                func_name="factorial",
                args=[5],
                target="test-serverless-mechanics-lambda",
            )
        )

        return v

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    main_node = Scheduler(store, logic_group=node_group)
    main_node.register(workflow, retry_policy=never())
    p: Promise[int] = main_node.run("test-serverless-mechanics", workflow)

    def factorial(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        time.sleep(0.5)  # This is not recommended. Just for testing.
        return n * (yield ctx.rfc(factorial, n - 1))

    node_delations: int = 0
    while not p.done():
        serverless_node = Scheduler(store, logic_group=lambda_node_group)
        serverless_node.register(factorial)
        serverless_node.wait_until_blocked(timeout=0.1)
        del serverless_node
        node_delations += 1

    assert p.done()
    assert p.result() == 120  # noqa: PLR2004
    assert node_delations > 1


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_fibonnaci_mechanics_awaiting() -> None:
    node_group = "test-fibonnaci-mechanics-awaiting"

    def fibonnaci(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n

        promise_id_n1 = f"fibonnaci-mechanics-awaiting-{n-1}"
        n1: int
        if n % randint(1, 10) == 0:  # noqa: S311
            n1 = yield ctx.rfc(fibonnaci, n - 1).with_options(
                promise_id=promise_id_n1, target=node_group
            )
        else:
            n1 = yield ctx.lfc(fibonnaci, n - 1).with_options(
                promise_id=promise_id_n1, retry_policy=never()
            )

        promise_id_n2 = f"fibonnaci-mechanics-awaiting-{n-2}"
        n2: int
        if n % randint(1, 10) == 0:  # noqa: S311
            n2 = yield ctx.rfc(fibonnaci, n - 2).with_options(
                promise_id=promise_id_n2, target=node_group
            )
        else:
            n2 = yield ctx.lfc(fibonnaci, n - 2).with_options(
                promise_id=promise_id_n2, retry_policy=never()
            )
        return n1 + n2

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    schedulers: list[Scheduler] = []
    for _ in range(randint(1, 5)):  # noqa: S311
        s = Scheduler(store, logic_group=node_group)
        s.register(fibonnaci, retry_policy=never())
        schedulers.append(s)
    n = randint(10, 30)  # noqa: S311

    p: Promise[int] = schedulers[0].run(
        f"fibonnaci-mechanics-awaiting-{n}", fibonnaci, n
    )
    assert p.result()


@pytest.mark.skip
@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_fibonnaci_mechanics_no_awaiting() -> None:
    node_group = "test-fibonnaci-mechanics-no-awaiting"

    def fibonnaci(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n

        promise_id_n1 = f"fibonnaci-mechanics-no-awaiting-{n-1}"
        pn1: Promise[int]
        if n % randint(1, 10) == 0:  # noqa: S311
            pn1 = yield ctx.rfi(fibonnaci, n - 1).with_options(
                promise_id=promise_id_n1, target=node_group
            )
        else:
            pn1 = yield ctx.lfi(fibonnaci, n - 1).with_options(
                promise_id=promise_id_n1, retry_policy=never()
            )

        promise_id_n2 = f"fibonnaci-mechanics-no-awaiting-{n-2}"
        pn2: Promise[int]
        if n % randint(1, 10) == 0:  # noqa: S311
            pn2 = yield ctx.rfi(fibonnaci, n - 2).with_options(
                promise_id=promise_id_n1, target=node_group
            )
        else:
            pn2 = yield ctx.lfi(fibonnaci, n - 2).with_options(
                promise_id=promise_id_n2, retry_policy=never()
            )

        n1 = yield pn1
        n2 = yield pn2
        return n1 + n2

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])
    schedulers: list[Scheduler] = []
    for _ in range(randint(1, 5)):  # noqa: S311
        s = Scheduler(store, logic_group=node_group)
        s.register(fibonnaci, retry_policy=never())
        schedulers.append(s)
    n = 8

    p: Promise[int] = schedulers[0].run(
        f"fibonnaci-mechanics-no-awaiting-{n}", fibonnaci, n
    )

    assert p.result() == 13  # noqa: PLR2004


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_trigger_on_other_process() -> None:
    node_group = "test-trigger-on-other-process"

    store = RemoteServer(url=os.environ["RESONATE_STORE_URL"])

    def factorial(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        return n * (yield ctx.rfc(factorial, n - 1))

    this_scheduler = Scheduler(store, logic_group=node_group)
    other_scheduler = Scheduler(store, logic_group=f"{node_group}-other")
    other_scheduler.register(factorial)
    p: Promise[int] = this_scheduler.trigger(
        f"{node_group}-factorial", "factorial", [5], target=f"{node_group}-other"
    )
    assert p.result() == 120  # noqa: PLR2004



from __future__ import annotations

from resonate.retry_policy import (
    calculate_total_possible_delay,
    constant,
    exponential,
    linear,
)


def test_exponential() -> None:
    policy = exponential(base_delay=1, factor=2, max_retries=3)
    expected_delays = (2, 4, 8)
    assert calculate_total_possible_delay(policy) == sum(expected_delays)
    assert policy.should_retry(attempt=1)
    assert policy.calculate_delay(attempt=1) == expected_delays[0]
    assert policy.should_retry(attempt=2)
    assert policy.calculate_delay(attempt=2) == expected_delays[1]
    assert policy.should_retry(attempt=3)
    assert policy.calculate_delay(attempt=3) == expected_delays[2]
    assert not policy.should_retry(attempt=4)


def test_linear() -> None:
    policy = linear(delay=2, max_retries=3)
    expected_delays = (2, 4, 6)
    assert calculate_total_possible_delay(policy) == sum(expected_delays)
    assert policy.should_retry(attempt=1)
    assert policy.calculate_delay(attempt=1) == expected_delays[0]
    assert policy.should_retry(attempt=2)
    assert policy.calculate_delay(attempt=2) == expected_delays[1]
    assert policy.should_retry(attempt=3)
    assert policy.calculate_delay(attempt=3) == expected_delays[2]
    assert not policy.should_retry(attempt=4)


def test_constant() -> None:
    policy = constant(delay=2, max_retries=3)
    expected_delays = (2, 2, 2)
    assert calculate_total_possible_delay(policy) == sum(expected_delays)
    assert policy.should_retry(attempt=1)
    assert policy.calculate_delay(attempt=1) == expected_delays[0]
    assert policy.should_retry(attempt=2)
    assert policy.calculate_delay(attempt=2) == expected_delays[1]
    assert policy.should_retry(attempt=3)
    assert policy.calculate_delay(attempt=3) == expected_delays[2]
    assert not policy.should_retry(attempt=4)



from __future__ import annotations

from collections import deque
from collections.abc import Coroutine, Generator
from typing import Any, Callable, Literal, TypeVar, Union

from typing_extensions import Concatenate, ParamSpec, TypeAlias

from resonate.actions import (
    LFC,
    LFI,
    RFC,
    RFI,
    All,
    AllSettled,
    DeferredInvocation,
    Race,
)
from resonate.batching import CmdBuffer
from resonate.commands import Command
from resonate.context import Context
from resonate.dataclasses import FnOrCoroutine, Runnable
from resonate.promise import Promise

T = TypeVar("T")
P = ParamSpec("P")


ExecutionUnit: TypeAlias = Union[Command, FnOrCoroutine]

Combinator: TypeAlias = Union[All, AllSettled, Race]

PromiseActions: TypeAlias = Union[Combinator, LFI, RFI]

Yieldable: TypeAlias = Union[LFC, Promise[Any], PromiseActions, DeferredInvocation, RFC]

DurableCoro: TypeAlias = Callable[Concatenate[Context, P], Generator[Yieldable, Any, T]]
DurableSyncFn: TypeAlias = Callable[Concatenate[Context, P], T]
DurableAsyncFn: TypeAlias = Callable[Concatenate[Context, P], Coroutine[Any, Any, T]]
DurableFn: TypeAlias = Union[DurableSyncFn[P, T], DurableAsyncFn[P, T]]
MockFn: TypeAlias = Callable[[], T]

Invokable: TypeAlias = Union[DurableCoro[P, Any], DurableFn[P, Any], Command]

AwaitingFor: TypeAlias = Literal["local", "remote"]
Runnables: TypeAlias = deque[tuple[Runnable[Any], bool]]

CommandHandlers: TypeAlias = dict[
    type[Command], Callable[[list[Any]], Union[list[Any], None]]
]
CommandHandlerQueues: TypeAlias = dict[
    type[Command], CmdBuffer[tuple[Promise[Any], Command]]
]


Headers: TypeAlias = Union[dict[str, str], None]
Tags: TypeAlias = Union[dict[str, str], None]
IdempotencyKey: TypeAlias = Union[str, None]
Data: TypeAlias = Union[str, None]
State: TypeAlias = Literal[
    "PENDING", "RESOLVED", "REJECTED", "REJECTED_CANCELED", "REJECTED_TIMEDOUT"
]


C = TypeVar("C", bound=Command)

CmdHandlerResult: TypeAlias = Union[list[Union[T, Exception]], T, None]
CmdHandler: TypeAlias = Callable[[Context, list[C]], CmdHandlerResult[Any]]



from __future__ import annotations

import contextlib
import heapq
import queue
from threading import Event, Thread
from typing import Generic, TypeVar, final

from resonate.time import now, ns_to_secs, secs_to_ns

T = TypeVar("T")


@final
class Queue(Generic[T]):
    def __init__(self, maxsize: int = 0) -> None:
        self._q = queue.Queue[T](maxsize=maxsize)

    def dequeue_batch(self, batch_size: int) -> list[T]:
        elements: list[T] = []
        with contextlib.suppress(queue.Empty):
            for _ in range(batch_size):
                elements.append(self._q.get_nowait())
                self._q.task_done()
        return elements

    def dequeue_all(self) -> list[T]:
        return self.dequeue_batch(self.qsize())

    def dequeue(self, timeout: float | None = None) -> T:
        qe = self._q.get(timeout=timeout)
        self._q.task_done()
        return qe

    def put_nowait(self, item: T) -> None:
        self._q.put_nowait(item)

    def put(self, item: T, *, block: bool = True, timeout: float | None = None) -> None:
        self._q.put(item, block, timeout)

    def qsize(self) -> int:
        return self._q.qsize()

    def get_nowait(self) -> T:
        return self._q.get_nowait()

    def task_done(self) -> None:
        self._q.task_done()

    def get(self, *, block: bool = True, timeout: float | None = None) -> T:
        return self._q.get(block, timeout)


@final
class DelayQueue(Generic[T]):
    def __init__(self, caller_event: Event | None, maxsize: int = 0) -> None:
        self._inq = Queue[tuple[T, float]](maxsize=maxsize)
        self._outq = Queue[T](maxsize=maxsize)
        self._delayed: list[tuple[float, int, T]] = []
        self._caller_event = caller_event
        self._continue_event = Event()

        self._worker_thread = Thread(target=self._run, daemon=True)
        self._worker_thread.start()

    def _next_release_time(self) -> float:
        return self._delayed[0][0]

    def _run(self) -> None:
        """Worker thread that processes the delayed queue."""
        while True:
            current_time = now()

            sqes = self._inq.dequeue_batch(self._inq.qsize())
            for idx, (item, delay) in enumerate(sqes):
                heapq.heappush(
                    self._delayed, (current_time + secs_to_ns(delay), idx, item)
                )

            # Release any items whose delay has expired
            while self._delayed and self._next_release_time() <= current_time:
                _, _, item = heapq.heappop(self._delayed)
                self._outq.put_nowait(item)  # Put the item in the consumer queue
                if self._caller_event:
                    self._caller_event.set()

            # Calculate the time to wait until the next item is ready

            wait_time: None | float = None
            if self._delayed:
                next_item_time = self._next_release_time()
                wait_time = max(0, next_item_time - current_time)

            self._continue_event.wait(
                ns_to_secs(wait_time) if wait_time is not None else wait_time
            )
            self._continue_event.clear()

    def dequeue_batch(self, batch_size: int) -> list[T]:
        return self._outq.dequeue_batch(batch_size)

    def dequeue(self, timeout: float | None = None) -> T:
        return self._outq.dequeue(timeout)

    def put_nowait(self, item: T, delay: float) -> None:
        self._inq.put_nowait((item, delay))
        self._continue_event.set()

    def items_in_delay(self) -> int:
        return len(self._delayed)

    def qsize(self) -> int:
        return self._outq.qsize()

    def dequeue_all(self) -> list[T]:
        return self._outq.dequeue_all()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic, TypeVar, final

from resonate.commands import Command

if TYPE_CHECKING:
    from resonate.promise import Promise

T = TypeVar("T")
C = TypeVar("C", bound=Command)


@final
@dataclass
class CmdBuffer(Generic[T]):
    max_length: int
    elements: list[list[T]] = field(init=False, default_factory=list)

    def append(self, e: T, /) -> None:
        if len(self.elements) == 0:
            self.elements.append([e])
        elif len(self.elements[-1]) < self.max_length:
            self.elements[-1].append(e)
        else:
            self.elements.append([e])

    def clear(self) -> None:
        self.elements.clear()


@final
class CommandBuffer(Generic[C]):
    def __init__(self, maxlen: int | None) -> None:
        self._maxlen = maxlen
        self.cmds: list[C] = []
        self.promises: list[Promise[Any]] = []

    def add(self, cmd: C, promise: Promise[Any]) -> None:
        assert len(self.cmds) == len(self.promises)
        assert not self.is_full()
        self.cmds.append(cmd)
        self.promises.append(promise)

    def is_full(self) -> bool:
        if self._maxlen is None:
            return False
        return self._maxlen == len(self.cmds)

    def is_empty(self) -> bool:
        return len(self.cmds) == 0

    def pop_all(self) -> tuple[list[C], list[Promise[Any]]]:
        cmds, self.cmds = self.cmds, []
        promises, self.promises = self.promises, []
        assert len(cmds) == len(promises)
        return cmds, promises



from __future__ import annotations

import random
from typing import TYPE_CHECKING, TypeVar, Union, final

from typing_extensions import ParamSpec, TypeAlias

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

T = TypeVar("T")
P = ParamSpec("P")

FloatOrInt: TypeAlias = Union[float, int]


@final
class Random:
    """Random class implementation that allows to enforce a prefixable behaviour."""

    def __init__(self, seed: int, prefix: list[FloatOrInt] | None = None) -> None:
        self.seed = seed
        self._prefix: list[FloatOrInt] = []
        if prefix is not None:
            self._prefix.extend(prefix)
        self._count = 0
        self._random = random.Random(seed)  # noqa: RUF100, S311

    def _take_number(
        self, fn: Callable[P, FloatOrInt], *args: P.args, **kwargs: P.kwargs
    ) -> FloatOrInt:
        self._count += 1
        if self._count <= len(self._prefix):
            return self._prefix[self._count - 1]

        random_number = fn(*args, **kwargs)
        self._prefix.append(random_number)
        return random_number

    def randint(self, a: int, b: int) -> int:
        return int(self._take_number(self._random.randint, a=a, b=b))

    def choice(self, seq: Sequence[T]) -> T:
        return seq[self.randint(0, len(seq) - 1)]

    def uniform(self, a: float, b: float) -> float:
        return self._take_number(self._random.uniform, a=a, b=b)

    def random(self) -> float:
        return self._take_number(self._random.random)

    def gauss(self, mu: float, sigma: float) -> float:
        return self._take_number(self._random.gauss, mu=mu, sigma=sigma)

    def randrange(self, start: int, stop: int | None = None, step: int = 1) -> int:
        return int(
            self._take_number(self._random.randrange, start=start, stop=stop, step=step)
        )

    def triangular(
        self, low: float = 0, high: float = 1, mode: float | None = None
    ) -> float:
        return self._take_number(self._random.triangular, low=low, high=high, mode=mode)

    def export(self) -> list[float]:
        return self._prefix



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypeVar, final

from typing_extensions import ParamSpec, Self

from resonate.commands import Command
from resonate.options import LOptions, ROptions
from resonate.retry_policy import never

if TYPE_CHECKING:
    from resonate.dataclasses import FnOrCoroutine
    from resonate.promise import Promise
    from resonate.retry_policy import RetryPolicy
    from resonate.typing import ExecutionUnit


P = ParamSpec("P")
T = TypeVar("T")


@final
@dataclass
class RFI:
    exec_unit: ExecutionUnit
    opts: ROptions = field(default=ROptions())

    def with_options(
        self, promise_id: str | None = None, target: str | None = None
    ) -> Self:
        assert not isinstance(
            self.exec_unit, Command
        ), "Options must be set on the command."
        self.opts = ROptions(promise_id=promise_id, target=target)
        return self


@final
@dataclass
class RFC:
    exec_unit: ExecutionUnit
    opts: ROptions = field(default=ROptions())

    def with_options(
        self, promise_id: str | None = None, target: str | None = None
    ) -> Self:
        assert not isinstance(
            self.exec_unit, Command
        ), "Options must be set on the command."
        self.opts = ROptions(promise_id=promise_id, target=target)
        return self

    def to_invocation(self) -> RFI:
        return RFI(self.exec_unit, self.opts)


@final
@dataclass
class LFC:
    exec_unit: ExecutionUnit
    opts: LOptions = field(default=LOptions())

    def with_options(
        self,
        promise_id: str | None = None,
        retry_policy: RetryPolicy | None = None,
        *,
        durable: bool = True,
    ) -> Self:
        if retry_policy is not None:
            assert not isinstance(
                self.exec_unit, Command
            ), "Retry policies on batching are set when registering command handlers."
        self.opts = LOptions(
            durable=durable, promise_id=promise_id, retry_policy=retry_policy
        )
        return self

    def to_invocation(self) -> LFI:
        return LFI(self.exec_unit, opts=self.opts)


@final
@dataclass
class DeferredInvocation:
    """
    Dataclass that contains all required information to do a
    deferred invocation.
    """

    promise_id: str
    coro: FnOrCoroutine
    opts: LOptions = field(default=LOptions())

    def with_options(self, *, retry_policy: RetryPolicy | None = None) -> Self:
        self.opts = LOptions(
            durable=True, promise_id=self.promise_id, retry_policy=retry_policy
        )
        return self


@final
@dataclass
class LFI:
    exec_unit: ExecutionUnit
    opts: LOptions = field(default=LOptions())

    def with_options(
        self,
        promise_id: str | None = None,
        retry_policy: RetryPolicy | None = None,
        *,
        durable: bool = True,
    ) -> Self:
        if retry_policy is not None:
            assert not isinstance(
                self.exec_unit, Command
            ), "Retry policies on batching are set when registering command handlers."
        self.opts = LOptions(
            durable=durable, promise_id=promise_id, retry_policy=retry_policy
        )
        return self


@final
class All:
    """
    A combinator that waits for all promises to complete.

    Attributes: promises (list[Promise[Any]]): A list of promises to be combined.
    """

    def __init__(self, promises: list[Promise[Any]]) -> None:
        self.opts = LOptions(retry_policy=never())
        self.promises = promises

    def with_options(
        self,
        *,
        durable: bool = True,
        promise_id: str | None = None,
        retry_policy: RetryPolicy | None = None,
    ) -> Self:
        """
        Set options for the combinator.

        Args: durable (bool): Whether the promise is durable. Defaults to True.
        promise_id (str | None): An optional identifier for the promise.
        retry_policy (RetryPolicy | None): An optional retry policy for the promise.

        Returns: Self: The combinator instance with updated options.
        """
        self.opts = LOptions(
            durable=durable,
            promise_id=promise_id,
            retry_policy=retry_policy if retry_policy is not None else never(),
        )
        return self


@final
class AllSettled:
    """
    A combinator that waits for all promises to complete and returns a list of results
    or Errors.

    Attributes: promises (list[Promise[Any]]): A list of promises to be combined.
    """

    def __init__(self, promises: list[Promise[Any]]) -> None:
        self.opts = LOptions(retry_policy=never())
        self.promises = promises

    def with_options(
        self,
        *,
        durable: bool = True,
        promise_id: str | None = None,
        retry_policy: RetryPolicy | None = None,
    ) -> Self:
        """
        Set options for the combinator.

        Args: durable (bool): Whether the promise is durable. Defaults to True.
        promise_id (str | None): An optional identifier for the promise.
        retry_policy (RetryPolicy | None): An optional retry policy for the promise.

        Returns: Self: The combinator instance with updated options.
        """
        self.opts = LOptions(
            durable=durable,
            promise_id=promise_id,
            retry_policy=retry_policy if retry_policy is not None else never(),
        )
        return self


@final
class Race:
    """
    A combinator that completes when any of the promises completes.

    Attributes: promises (list[Promise[Any]]): A list of promises to race.
    """

    def __init__(self, promises: list[Promise[Any]]) -> None:
        assert (
            len(promises) > 0
        ), "Race combinator requires a non empty list of promises"
        self.opts = LOptions(retry_policy=never())
        self.promises = promises

    def with_options(
        self,
        *,
        durable: bool = True,
        promise_id: str | None = None,
        retry_policy: RetryPolicy | None = None,
    ) -> Self:
        """
        Set options for the combinator.

        Args: durable (bool): Whether the promise is durable. Defaults to True.
        promise_id (str | None): An optional identifier for the promise.
        retry_policy (RetryPolicy | None): An optional retry policy for the promise.

        Returns: Self: The combinator instance with updated options.
        """
        self.opts = LOptions(
            durable=durable,
            promise_id=promise_id,
            retry_policy=retry_policy if retry_policy is not None else never(),
        )
        return self



from __future__ import annotations

from concurrent.futures import Future
from typing import TYPE_CHECKING, Any, Generic, TypeVar, final

from typing_extensions import assert_never

from resonate.actions import (
    LFI,
    RFI,
    All,
    AllSettled,
    Race,
)
from resonate.result import Err, Ok, Result

if TYPE_CHECKING:
    from resonate.typing import PromiseActions

T = TypeVar("T")


@final
class Promise(Generic[T]):
    def __init__(
        self,
        promise_id: str,
        action: PromiseActions,
        parent_promise: Promise[Any] | None,
    ) -> None:
        self.promise_id = promise_id
        self.f = Future[T]()
        self._next_child_num = 1
        self.parent_promise = parent_promise
        self.children_promises: list[Promise[Any]] = []
        self.leaf_promises = set[Promise[Any]]()
        self._is_partition_root = self.parent_promise is None
        self._marked_as_partition_root = False

        self.action = action
        if isinstance(action, (LFI, All, AllSettled, Race)):
            self.durable = action.opts.durable
        elif isinstance(action, RFI):
            self.durable = True
        else:
            assert_never(action)

    def is_partition_root(self) -> bool:
        return self._is_partition_root or self._marked_as_partition_root

    def is_marked_as_partition_root(self) -> bool:
        return self._marked_as_partition_root

    def unmark_as_partition_root(self) -> None:
        assert (
            self._marked_as_partition_root
        ), "Promise is already marked as partition root"
        self._marked_as_partition_root = not self._marked_as_partition_root

    def mark_as_partition_root(self) -> None:
        assert not self._is_partition_root, "Promise is already a partition root"
        assert (
            not self._marked_as_partition_root
        ), "Promise is already marked as partition root"
        self._marked_as_partition_root = not self._marked_as_partition_root

    def result(self, timeout: float | None = None) -> T:
        return self.f.result(timeout=timeout)

    def safe_result(self, timeout: float | None = None) -> Result[T, Exception]:
        res: Result[T, Exception]
        try:
            res = Ok(self.f.result(timeout=timeout))
        except Exception as e:  # noqa: BLE001
            res = Err(e)
        return res

    def set_result(self, result: Result[T, Exception]) -> None:
        if isinstance(result, Ok):
            self.f.set_result(result.unwrap())
        elif isinstance(result, Err):
            self.f.set_exception(result.err())
        else:
            assert_never(result)

    def done(self) -> bool:
        return self.f.done()

    def success(self) -> bool:
        if not self.done():
            return False
        try:
            self.result()
        except Exception:  # noqa: BLE001
            return False
        else:
            return True

    def failure(self) -> bool:
        return not self.success()

    def parent_promise_id(self) -> str | None:
        return (
            self.parent_promise.promise_id if self.parent_promise is not None else None
        )

    def root(self) -> Promise[Any]:
        maybe_root = self
        while True:
            if maybe_root.parent_promise is None:
                return maybe_root
            maybe_root = maybe_root.parent_promise

    def partition_root(self) -> Promise[Any]:
        maybe_root = self
        while True:
            if maybe_root.is_partition_root():
                return maybe_root
            assert maybe_root.parent_promise is not None
            maybe_root = maybe_root.parent_promise

    def child_name(self) -> str:
        return f"{self.promise_id}.{self._next_child_num}"

    def child_promise(
        self,
        promise_id: str,
        action: PromiseActions,
    ) -> Promise[Any]:
        child = Promise[Any](
            promise_id=promise_id,
            action=action,
            parent_promise=self,
        )

        self.children_promises.append(child)
        self.leaf_promises.add(child)
        root_promise = self.root()
        if root_promise is not None:
            root_promise.leaf_promises.discard(self)
            root_promise.leaf_promises.add(child)

        self._next_child_num += 1
        return child


def all_promises_are_done(promises: list[Promise[Any]]) -> bool:
    return all(p.done() for p in promises)


def any_promise_is_done(promises: list[Promise[Any]]) -> bool:
    return any(p.done() for p in promises)


def get_first_error_if_any(promises: list[Promise[Any]]) -> Err[Exception] | None:
    for p in promises:
        if p.success():
            continue
        error = p.safe_result()
        assert isinstance(error, Err)
        return error
    return None
